import io
import math
from zipfile import ZipFile
from urllib.request import urlopen
import pandas as pd
import numpy as np
import bciavm
from bciavm.core.config import your_bucket, target, TABLE_DIRECTORY, WGS84_a, WGS84_b, box_in_km
from sklearn.model_selection import ShuffleSplit, StratifiedShuffleSplit

target = 'Price_p'

def deg2rad(degrees):
    return math.pi * degrees / 180.0

def rad2deg(radians):
    return 180.0 * radians / math.pi

def get_latmin(lat):
    halfSide = 1000 * box_in_km

    # Radius of Earth at given latitude
    An = WGS84_a * WGS84_a * math.cos(lat)
    Bn = WGS84_b * WGS84_b * math.sin(lat)
    Ad = WGS84_a * math.cos(lat)
    Bd = WGS84_b * math.sin(lat)
    radius = math.sqrt((An * An + Bn * Bn) / (Ad * Ad + Bd * Bd))

    # Radius of the parallel at given latitude
    pradius = radius * math.cos(lat)
    latMin = lat - halfSide / radius
    latMax = lat + halfSide / radius
    return latMin

def get_latmax(lat):
    halfSide = 1000 * box_in_km

    # Radius of Earth at given latitude
    An = WGS84_a * WGS84_a * math.cos(lat)
    Bn = WGS84_b * WGS84_b * math.sin(lat)
    Ad = WGS84_a * math.cos(lat)
    Bd = WGS84_b * math.sin(lat)
    radius = math.sqrt((An * An + Bn * Bn) / (Ad * Ad + Bd * Bd))

    # Radius of the parallel at given latitude
    pradius = radius * math.cos(lat)
    latMin = lat - halfSide / radius
    latMax = lat + halfSide / radius
    return latMax


def get_pradius(lat):
    halfSide = 1000 * box_in_km

    # Radius of Earth at given latitude
    An = WGS84_a * WGS84_a * math.cos(lat)
    Bn = WGS84_b * WGS84_b * math.sin(lat)
    Ad = WGS84_a * math.cos(lat)
    Bd = WGS84_b * math.sin(lat)
    radius = math.sqrt((An * An + Bn * Bn) / (Ad * Ad + Bd * Bd))

    # Radius of the parallel at given latitude
    pradius = radius * math.cos(lat)
    return pradius

def get_postcodeOutcode_from_postcode(postcode):
    return postcode.split()[0]

def get_postcode_from_address(string):
    return string.split(' ')[-2] + ' ' + string.split(' ')[-1]

def get_postcodeArea_from_outcode(postcodeArea):
    if postcodeArea[1].isnumeric():
        return postcodeArea[0]
    else:
        return postcodeArea[0:2]

def ReadParquetFile(bucketName, fileLocation):
    df = pd.DataFrame()
    prefix_objs = your_bucket.objects.filter(Prefix=fileLocation)
    for s3_file in prefix_objs:
        obj = s3_file.get()
        df = df.append(pd.read_parquet(io.BytesIO(obj['Body'].read())))
    return df

def download_postcodes(path=TABLE_DIRECTORY + 'data/raw/ukpostcodes'):
    zipurl = 'https://www.freemaptools.com/download/full-postcodes/ukpostcodes.zip'

    # Download the file from the URL
    zipresp = urlopen(zipurl)

    # Create a new file on the hard drive
    tempzip = open("/tmp/tempfile.zip", "wb")

    # Write the contents of the downloaded file into the new file
    tempzip.write(zipresp.read())

    # Close the newly-created file
    tempzip.close()

    # Re-open the newly-created file with ZipFile()
    zf = ZipFile("/tmp/tempfile.zip")

    # Extract its contents into <extraction_path>
    # note that extractall will automatically create the path
    zf.extractall(path)

    # close the ZipFile instance
    zf.close()

def drop_outliers(data):
    upper_outliers, lower_outliers = np.quantile(data['Price_p'], 0.99), np.quantile(data['Price_p'], 0.01)
    data = data[(data['Price_p'] <= upper_outliers) & (data['Price_p'] >= lower_outliers)]
    data = data[
        (data['Price_p'] <= np.quantile(data['Price_p'], 0.99)
         ) & (data['Price_p'] >= 10000)]
    data = data[data['NUMBER_HABITABLE_ROOMS_e'].astype(float) < 10]
    data = data[data['FLAT_STOREY_COUNT_e'].astype(float).fillna(0) < 20]
    data = data[data['TOTAL_FLOOR_AREA_e'].astype(float) < 800]
    data = data[data['TOTAL_FLOOR_AREA_e'].astype(float) > 0]
    return data


def preprocess_data(data, drop_outlier=True, target='Price_p', test_size=0.15, split_data=True):

    if drop_outlier:
        data = drop_outliers(data)

    keep_cols = [x for x in data.columns if '_e' in x or x in ['Longitude_m', 'Latitude_m', 'Postcode']]
    if target in data.columns:
        keep_cols.append(target)

    try:
        match_types = ['1. Address Matched', '2. Address Matched No Spec', '3. No in Address Matched']
        data = data[data['TypeOfMatching_m'].isin(match_types)]
    except: pass

    data = data[keep_cols]

    try:
        data['POSTCODE'] = data['FULLADRESS_e'].apply(get_postcode_from_address)
        data['POSTCODE_OUTCODE'] = data['Postcode'].apply(get_postcodeOutcode_from_postcode)
        data['POSTCODE_AREA'] = data['POSTCODE_OUTCODE'].apply(get_postcodeArea_from_outcode)
    except: pass

    # drop outliers, convert floor level to integers
    data['Rooms'] = (data['NUMBER_HABITABLE_ROOMS_e'].astype(float) + data['NUMBER_HEATED_ROOMS_e'].astype(
        float)) / float(2)

    data['PROPERTY_TYPE_e'] = data['PROPERTY_TYPE_e'].astype(str).replace('nan', np.nan).fillna(
        'No PROPERTY_TYPE_e').astype(str)
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('NO DATA!', '0')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('NODATA!', '0')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('Basement', '-1')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('Ground', '0')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('1st', '1')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('2nd', '2')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('3rd', '3')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('4th', '4')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('5th', '5')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('6th', '6')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('9th', '9')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('11th', '11')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('12th', '12')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('13th', '13')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('14th', '14')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('15th', '15')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('16th', '16')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('17th', '17')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('18th', '18')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('19th', '19')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('20th', '20')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('21st or above', '21')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('top floor', '22')
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].replace('mid floor', '3')
    floor_levels = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17','18', '19', '20', '21', '22', '-1']
    floor_levels = [x for x in data['FLOOR_LEVEL_e'].values if x not in floor_levels]
    data['FLOOR_LEVEL_e'] = np.where(data['FLOOR_LEVEL_e'].isin(floor_levels), np.nan, data['FLOOR_LEVEL_e']).astype(
        float)
    data['NUMBER_HEATED_ROOMS_e'] = data['NUMBER_HEATED_ROOMS_e'].astype(str).replace('nan', np.nan).fillna('0').astype(
        float)
    if target in data.columns:
        data = data[['POSTCODE',
                     'POSTCODE_OUTCODE',
                     'POSTCODE_AREA',
                     'POSTTOWN_e',
                     'PROPERTY_TYPE_e',
                     'TOTAL_FLOOR_AREA_e',
                     'NUMBER_HEATED_ROOMS_e',
                     'FLOOR_LEVEL_e',
                     'Latitude_m',
                     'Longitude_m',
                     target]]
    else:
        data = data[['POSTCODE',
                     'POSTCODE_OUTCODE',
                     'POSTCODE_AREA',
                     'POSTTOWN_e',
                     'PROPERTY_TYPE_e',
                     'TOTAL_FLOOR_AREA_e',
                     'NUMBER_HEATED_ROOMS_e',
                     'FLOOR_LEVEL_e',
                     'Latitude_m',
                     'Longitude_m']]

    data = data.reset_index()
    data = data.drop('index', axis=1)
    data = data.reset_index()
    data = data.rename({'index': 'unit_indx'}, axis=1)
    data = data.reset_index()
    data = data.drop('index', axis=1)
    data['TOTAL_FLOOR_AREA_e'] = data['TOTAL_FLOOR_AREA_e'].astype(float)
    data['FLOOR_LEVEL_e'] = data['FLOOR_LEVEL_e'].astype(float)
    data['NUMBER_HEATED_ROOMS_e'] = data['NUMBER_HEATED_ROOMS_e'].astype(float)

    if target in data.columns:
        data[target] = data[target].astype(float)

    if not split_data:
        return data

    X = data.drop('POSTCODE_AREA', axis=1)
    y = data['POSTCODE_AREA']
    X_train, X_holdout, y_train, y_holdout = bciavm.preprocessing.utils.split_data(X, y, problem_type='multiclass',
                                                                                   test_size=test_size)

    validate = X_holdout.reset_index().drop('index', axis=1)
    validate['POSTCODE_AREA'] = y_holdout.reset_index()['POSTCODE_AREA']

    train = X_train.reset_index().drop('index', axis=1)
    train['POSTCODE_AREA'] = y_train.reset_index()['POSTCODE_AREA']

    # additional processing to create train+test sets
    data = train
    data = data.dropna(subset=[target, 'TOTAL_FLOOR_AREA_e'])
    X, y = data, data[target]
    X, y = bciavm.preprocessing.utils.drop_nan_target_rows(X, y)

    # create a new feature upon which we will split the data into train+test sets
    X['key'] = X['POSTCODE_AREA'].astype(str) + '_' + X['PROPERTY_TYPE_e'].astype(str)
    vals = X['key'].value_counts().reset_index()
    vals = vals[vals['key'] > 5]
    X = X[X['key'].isin(list(vals['index'].values))]

    # split into train+test sets
    data_splitter = StratifiedShuffleSplit(n_splits=1, test_size=test_size, random_state=0)
    y = X[target]
    train, test = next(data_splitter.split(X, X['key']))
    X = X.drop('key', axis=1)
    X_train = X.iloc[train]
    X_test = X.iloc[test]
    y_train = y.iloc[train]
    y_test = y.iloc[test]
    X_train = X_train.drop(target, axis=1)
    X_test = X_test.drop(target, axis=1)

    return X_train, X_test, y_train, y_test

def merge_train_w_lookup_table(train, lookup_table):
    if 'PROPERTY_TYPE_e__' in lookup_table.columns:
        train = train.merge(lookup_table.rename({'PROPERTY_TYPE_e__': 'PROPERTY_TYPE_e'}, axis=1),
                            on=['POSTCODE', 'PROPERTY_TYPE_e'])
    else:
        train = train.merge(lookup_table, on=['POSTCODE', 'PROPERTY_TYPE_e'])

    new_X_train = pd.DataFrame({})

    for col in ['FLOOR_LEVEL_e', 'NUMBER_HEATED_ROOMS_e', 'TOTAL_FLOOR_AREA_e', 'Price_p']:
        agg_cols = [c for c in train.columns if col in c and c != col]
        for agg_col in agg_cols:
            if col != 'Price_p':
                new_X_train[col + '_minus_' + agg_col] = train[col] - train[agg_col]
            else:
                new_X_train[agg_col] = train[agg_col]

    new_X_train['Latitude_m'] = train['Latitude_m']
    new_X_train['Longitude_m'] = train['Longitude_m']
    new_X_train['PROPERTY_TYPE_e'] = train['PROPERTY_TYPE_e']
    new_X_train['POSTTOWN_e'] = train['POSTTOWN_e']
    new_X_train['POSTCODE'] = train['POSTCODE']
    new_X_train['POSTCODE_AREA'] = train['POSTCODE_AREA']
    new_X_train['density_count'] = train['density_count']
    new_X_train[target] = train[target]
    return new_X_train


import numpy as np
from sklearn.ensemble import RandomForestClassifier as SKRandomForestClassifier
from sklearn.feature_selection import SelectFromModel as SkSelect
from skopt.space import Real

from .feature_selector import FeatureSelector


class RFClassifierSelectFromModel(FeatureSelector):
    """Selects top features based on importance weights using a Random Forest classifier."""
    name = 'RF Classifier Select From Model'
    hyperparameter_ranges = {
        "percent_features": Real(.01, 1),
        "threshold": ['mean', -np.inf]
    }

    def __init__(self, number_features=None, n_estimators=10, max_depth=None,
                 percent_features=0.5, threshold=-np.inf, n_jobs=-1, random_seed=0, **kwargs):
        parameters = {"number_features": number_features,
                      "n_estimators": n_estimators,
                      "max_depth": max_depth,
                      "percent_features": percent_features,
                      "threshold": threshold,
                      "n_jobs": n_jobs}
        parameters.update(kwargs)

        estimator = SKRandomForestClassifier(random_state=random_seed,
                                             n_estimators=n_estimators,
                                             max_depth=max_depth,
                                             n_jobs=n_jobs)
        max_features = max(1, int(percent_features * number_features)) if number_features else None
        feature_selection = SkSelect(estimator=estimator,
                                     max_features=max_features,
                                     threshold=threshold,
                                     **kwargs)
        super().__init__(parameters=parameters,
                         component_obj=feature_selection,
                         random_seed=random_seed)



from bciavm.pipelines.components.transformers import Transformer
from bciavm.pipelines.components.transformers.imputers.simple_imputer import (
    SimpleImputer
)
from bciavm.utils import (
    _retain_custom_types_and_initalize_woodwork,
    infer_feature_types
)


class PerColumnImputer(Transformer):
    """Imputes missing data according to a specified imputation strategy per column"""
    name = 'Per Column Imputer'
    hyperparameter_ranges = {}

    def __init__(self, impute_strategies=None, default_impute_strategy="most_frequent", random_seed=0, **kwargs):
        """Initializes a transformer that imputes missing data according to the specified imputation strategy per column."

        Arguments:
            impute_strategies (dict): Column and {"impute_strategy": strategy, "fill_value":value} pairings.
                Valid values for impute strategy include "mean", "median", "most_frequent", "constant" for numerical data,
                and "most_frequent", "constant" for object data types. Defaults to "most_frequent" for all columns.

                When impute_strategy == "constant", fill_value is used to replace missing data.
                Defaults to 0 when imputing numerical data and "missing_value" for strings or object data types.

            default_impute_strategy (str): Impute strategy to fall back on when none is provided for a certain column.
                Valid values include "mean", "median", "most_frequent", "constant" for numerical data,
                and "most_frequent", "constant" for object data types. Defaults to "most_frequent"

            random_seed (int): Seed for the random number generator. Defaults to 0.
        """
        parameters = {"impute_strategies": impute_strategies,
                      "default_impute_strategy": default_impute_strategy}
        self.imputers = None
        self.default_impute_strategy = default_impute_strategy
        self.impute_strategies = impute_strategies or dict()

        if not isinstance(self.impute_strategies, dict):
            raise ValueError("`impute_strategies` is not a dictionary. Please provide in Column and {`impute_strategy`: strategy, `fill_value`:value} pairs. ")

        super().__init__(parameters=parameters,
                         component_obj=None,
                         random_seed=random_seed)

    def fit(self, X, y=None):
        """Fits imputers on input data

        Arguments:
            X (ww.DataTable, pd.DataFrame or np.ndarray): The input training data of shape [n_samples, n_features] to fit.
            y (ww.DataColumn, pd.Series, optional): The target training data of length [n_samples]. Ignored.

        Returns:
            self
        """
        X = infer_feature_types(X)
        self.imputers = dict()
        for column in X.columns:
            strategy_dict = self.impute_strategies.get(column, dict())
            strategy = strategy_dict.get('impute_strategy', self.default_impute_strategy)
            fill_value = strategy_dict.get('fill_value', None)
            self.imputers[column] = SimpleImputer(impute_strategy=strategy, fill_value=fill_value)

        for column, imputer in self.imputers.items():
            imputer.fit(X[[column]])

        return self

    def transform(self, X, y=None):
        """Transforms input data by imputing missing values.

        Arguments:
            X (ww.DataTable, pd.DataFrame or np.ndarray): The input training data of shape [n_samples, n_features] to transform.
            y (ww.DataColumn, pd.Series, optional): The target training data of length [n_samples]. Ignored.

        Returns:
            ww.DataTable: Transformed X
        """
        X_ww = infer_feature_types(X)
        import copy
        original_ltypes = copy.deepcopy(X_ww.ww.logical_types)
        cols_to_drop = []
        for column, imputer in self.imputers.items():
            transformed = imputer.transform(X_ww[[column]])
            if transformed.empty:
                cols_to_drop.append(column)
            else:
                X_ww.ww[column] = transformed[column]
        X_t = X_ww.ww.drop(cols_to_drop)
        return _retain_custom_types_and_initalize_woodwork(original_ltypes, X_t)



from sklearn.linear_model import LinearRegression as SKLinearRegression
from bciavm.model_family import ModelFamily
from bciavm.pipelines.components.estimators import Estimator
from bciavm.problem_types import ProblemTypes


class LinearRegressor(Estimator):
    """Linear Regressor."""
    name = "Linear Regressor"
    hyperparameter_ranges = {
        'fit_intercept': [True, False],
        'normalize': [True, False]
    }
    model_family = ModelFamily.LINEAR_MODEL
    supported_problem_types = [ProblemTypes.REGRESSION, ProblemTypes.TIME_SERIES_REGRESSION]

    def __init__(self, fit_intercept=True, normalize=False, n_jobs=-1, random_seed=0, **kwargs):
        parameters = {
            'fit_intercept': fit_intercept,
            'normalize': normalize,
            'n_jobs': n_jobs
        }
        parameters.update(kwargs)
        linear_regressor = SKLinearRegression(**parameters)
        super().__init__(parameters=parameters,
                         component_obj=linear_regressor,
                         random_seed=random_seed)

    @property
    def feature_importance(self):
        return self._component_obj.coef_



from abc import ABC, abstractmethod


class SamplerBase(ABC):
    """Base class for all custom samplers"""

    def __init__(self, random_seed=0):
        self.random_seed = random_seed

    @abstractmethod
    def fit_resample(self, X, y):
        """Resample the input data with this sampling strategy.

        Arguments:
            X (pd.DataFrame): Training data to fit and resample
            y (pd.Series): Training data targets to fit and resample

        Returns:
            Tuple(pd.DataFrame, pd.Series) or list: resampled X and y data for oversampling or indices to keep for undersampling
        """



import pandas as pd
from sklearn.impute import SimpleImputer as SkImputer

from bciavm.pipelines.components.transformers import Transformer
from bciavm.utils import (
    _retain_custom_types_and_initalize_woodwork,
    infer_feature_types
)
from bciavm import Boolean
import woodwork as ww

class SimpleImputer(Transformer):
    """Imputes missing data according to a specified imputation strategy."""
    name = 'Simple Imputer'
    hyperparameter_ranges = {"impute_strategy": ["mean", "median", "most_frequent"]}

    def __init__(self, impute_strategy="most_frequent", fill_value=None, random_seed=0, **kwargs):
        """Initalizes an transformer that imputes missing data according to the specified imputation strategy."

        Arguments:
            impute_strategy (string): Impute strategy to use. Valid values include "mean", "median", "most_frequent", "constant" for
               numerical data, and "most_frequent", "constant" for object data types.
            fill_value (string): When impute_strategy == "constant", fill_value is used to replace missing data.
               Defaults to 0 when imputing numerical data and "missing_value" for strings or object data types.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        """
        parameters = {"impute_strategy": impute_strategy,
                      "fill_value": fill_value}
        parameters.update(kwargs)
        imputer = SkImputer(strategy=impute_strategy,
                            fill_value=fill_value,
                            **kwargs)
        self._all_null_cols = None
        super().__init__(parameters=parameters,
                         component_obj=imputer,
                         random_seed=random_seed)

    def fit(self, X, y=None):
        """Fits imputer to data. 'None' values are converted to np.nan before imputation and are
            treated as the same.

        Arguments:
            X (ww.DataTable, pd.DataFrame or np.ndarray): the input training data of shape [n_samples, n_features]
            y (ww.DataColumn, pd.Series, optional): the target training data of length [n_samples]

        Returns:
            self
        """
        X = infer_feature_types(X)

        # Convert all bool dtypes to category for fitting
        if (X.dtypes == bool).all():
            X = X.astype('category')

        self._component_obj.fit(X, y)
        self._all_null_cols = set(X.columns) - set(X.dropna(axis=1, how='all').columns)
        return self

    def transform(self, X, y=None):
        """Transforms input by imputing missing values. 'None' and np.nan values are treated as the same.

        Arguments:
            X (ww.DataTable, pd.DataFrame): Data to transform
            y (ww.DataColumn, pd.Series, optional): Ignored.

        Returns:
            ww.DataTable: Transformed X
        """
        X_ww = infer_feature_types(X)

        # Return early since bool dtype doesn't support nans and sklearn errors if all cols are bool
        if (X_ww.dtypes == bool).all():
            return X_ww

        X_null_dropped = X_ww.ww.drop(self._all_null_cols)
        X_t = self._component_obj.transform(X_ww)

        # Need this for test_simple_imputer_multitype_with_one_bool
        X_t = pd.DataFrame(X_t, columns=X_null_dropped.columns)
        if not X_null_dropped.empty:
            X_t.index = X_null_dropped.index

        return _retain_custom_types_and_initalize_woodwork(X_ww.ww.logical_types, X_t)

    def fit_transform(self, X, y=None):
        """Fits on X and transforms X

        Arguments:
            X (ww.DataTable, pd.DataFrame): Data to fit and transform
            y (ww.DataColumn, pd.Series, optional): Target data.

        Returns:
            ww.DataTable: Transformed X
        """
        return self.fit(X, y).transform(X, y)



from .logger import get_logger, log_subtitle, log_title
from .gen_utils import (
    classproperty,
    import_or_raise,
    convert_to_seconds,
    get_random_state,
    get_random_seed,
    SEED_BOUNDS,
    jupyter_check,
    safe_repr,
    drop_rows_with_nans,
    pad_with_nans,
    _get_rows_without_nans,
    save_plot,
    is_all_numeric,
    get_importable_subclasses,
    _rename_column_names_to_numeric,
    deprecate_arg,
    )
from .cli_utils import (
    get_bciavm_root,
    get_installed_packages,
    get_sys_info,
    print_deps,
    print_info,
    print_sys_info,
    )
from .woodwork_utils import (
    _retain_custom_types_and_initalize_woodwork,
    infer_feature_types,
    _convert_numeric_dataset_pandas,
    )




import pandas as pd
from pandas.api.types import is_numeric_dtype

from .problem_types import ProblemTypes


def handle_problem_types(problem_type):
    """Handles problem_type by either returning the ProblemTypes or converting from a str.

    Arguments:
        problem_type (str or ProblemTypes): Problem type that needs to be handled

    Returns:
        ProblemTypes
    """
    if isinstance(problem_type, str):
        try:
            tpe = ProblemTypes._all_values[problem_type.upper()]
        except KeyError:
            raise KeyError('Problem type \'{}\' does not exist'.format(problem_type))
        return tpe
    if isinstance(problem_type, ProblemTypes):
        return problem_type
    raise ValueError('`handle_problem_types` was not passed a str or ProblemTypes object')


def detect_problem_type(y):
    """Determine the type of problem is being solved based on the targets (binary vs multiclass classification, regression)
        Ignores missing and null data

    Arguments:
        y (pd.Series): the target labels to predict

    Returns:
        ProblemType: ProblemType Enum

    Example:
        >>> y = pd.Series([0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 1])
        >>> problem_type = detect_problem_type(y)
        >>> assert problem_type == ProblemTypes.BINARY
    """
    y = pd.Series(y).dropna()
    num_classes = y.nunique()
    if num_classes < 2:
        raise ValueError("Less than 2 classes detected! Target unusable for modeling")
    if num_classes == 2:
        return ProblemTypes.BINARY
    if is_numeric_dtype(y.dtype):
        if (num_classes > 10):
            return ProblemTypes.REGRESSION
    return ProblemTypes.MULTICLASS


def is_regression(problem_type):
    """Determines if the provided problem_type is a regression problem type

    Arguments:
        problem_type (str or ProblemTypes): type of supervised learning problem. See bciavm.problem_types.ProblemType.all_problem_types for a full list.

    Returns:
        bool: Whether or not the provided problem_type is a regression problem type.
"""
    return handle_problem_types(problem_type) in [ProblemTypes.REGRESSION, ProblemTypes.TIME_SERIES_REGRESSION]


def is_binary(problem_type):
    """Determines if the provided problem_type is a binary classification problem type

    Arguments:
        problem_type (str or ProblemTypes): type of supervised learning problem. See bciavm.problem_types.ProblemType.all_problem_types for a full list.

    Returns:
        bool: Whether or not the provided problem_type is a binary classification problem type.
"""
    return handle_problem_types(problem_type) in [ProblemTypes.BINARY, ProblemTypes.TIME_SERIES_BINARY]


def is_multiclass(problem_type):
    """Determines if the provided problem_type is a multiclass classification problem type

    Arguments:
        problem_type (str or ProblemTypes): type of supervised learning problem. See bciavm.problem_types.ProblemType.all_problem_types for a full list.

    Returns:
        bool: Whether or not the provided problem_type is a multiclass classification problem type.
"""
    return handle_problem_types(problem_type) in [ProblemTypes.MULTICLASS, ProblemTypes.TIME_SERIES_MULTICLASS]


def is_classification(problem_type):
    """Determines if the provided problem_type is a classification problem type

    Arguments:
        problem_type (str or ProblemTypes): type of supervised learning problem. See bciavm.problem_types.ProblemType.all_problem_types for a full list.

    Returns:
        bool: Whether or not the provided problem_type is a classification problem type.
"""
    return is_binary(problem_type) or is_multiclass(problem_type)


def is_time_series(problem_type):
    """Determines if the provided problem_type is a time series problem type

    Arguments:
        problem_type (str or ProblemTypes): type of supervised learning problem. See bciavm.problem_types.ProblemType.all_problem_types for a full list.

    Returns:
        bool: Whether or not the provided problem_type is a time series problem type.
"""
    return handle_problem_types(problem_type) in [ProblemTypes.TIME_SERIES_BINARY,
                                                  ProblemTypes.TIME_SERIES_MULTICLASS,
                                                  ProblemTypes.TIME_SERIES_REGRESSION]



from bciavm.pipelines import RegressionPipeline


class BaselineRegressionPipeline(RegressionPipeline):
    """Baseline Pipeline for regression problems."""
    _name = "Baseline Regression Pipeline"
    component_graph = ["Baseline Regressor"]


class MeanBaselineRegressionPipeline(RegressionPipeline):
    """Baseline Pipeline for regression problems."""
    _name = "Mean Baseline Regression Pipeline"
    component_graph = ["Baseline Regressor"]
    custom_hyperparameters = {"strategy": ["mean"]}



import pandas as pd

from bciavm.pipelines.pipeline_meta import TimeSeriesPipelineBaseMeta
from bciavm.pipelines.regression_pipeline import RegressionPipeline
from bciavm.problem_types import ProblemTypes
from bciavm.utils import (
    drop_rows_with_nans,
    infer_feature_types,
    pad_with_nans
)


class TimeSeriesRegressionPipeline(RegressionPipeline, metaclass=TimeSeriesPipelineBaseMeta):
    """Pipeline base class for time series regression problems."""

    problem_type = ProblemTypes.TIME_SERIES_REGRESSION

    def __init__(self, parameters, random_seed=0):
        """Machine learning pipeline for time series regression problems made out of transformers and a classifier.

        Required Class Variables:
            component_graph (list): List of components in order. Accepts strings or ComponentBase subclasses in the list

        Arguments:
            parameters (dict): Dictionary with component names as keys and dictionary of that component's parameters as values.
                 An empty dictionary {} implies using all default values for component parameters. Pipeline-level
                 parameters such as gap and max_delay must be specified with the "pipeline" key. For example:
                 Pipeline(parameters={"pipeline": {"max_delay": 4, "gap": 2}}).
            random_seed (int): Seed for the random number generator. Defaults to 0.
        """
        if "pipeline" not in parameters:
            raise ValueError("gap and max_delay parameters cannot be omitted from the parameters dict. "
                             "Please specify them as a dictionary with the key 'pipeline'.")
        pipeline_params = parameters["pipeline"]
        self.gap = pipeline_params['gap']
        self.max_delay = pipeline_params['max_delay']
        super().__init__(parameters, random_seed=random_seed)

    def fit(self, X, y):
        """Fit a time series regression pipeline.

        Arguments:
            X (ww.DataTable, pd.DataFrame or np.ndarray): The input training data of shape [n_samples, n_features]
            y (ww.DataColumn, pd.Series, np.ndarray): The target training targets of length [n_samples]

        Returns:
            self
        """
        if X is None:
            X = pd.DataFrame()

        X = infer_feature_types(X)
        y = infer_feature_types(y)
        X_t = self._compute_features_during_fit(X, y)

        y_shifted = y.shift(-self.gap)
        X_t, y_shifted = drop_rows_with_nans(X_t, y_shifted)
        self.estimator.fit(X_t, y_shifted)
        self.input_feature_names = self._component_graph.input_feature_names

        return self

    def predict(self, X, y=None, objective=None):
        """Make predictions using selected features.

        Arguments:
            X (ww.DataTable, pd.DataFrame, or np.ndarray): Data of shape [n_samples, n_features]
            y (ww.DataColumn, pd.Series, np.ndarray, None): The target training targets of length [n_samples]
            objective (Object or string): The objective to use to make predictions

        Returns:
            ww.DataColumn: Predicted values.
        """
        if X is None:
            X = pd.DataFrame()
        X = infer_feature_types(X)
        y = infer_feature_types(y)
        features = self.compute_estimator_features(X, y)
        features_no_nan, y = drop_rows_with_nans(features, y)
        y_arg = None
        if self.estimator.predict_uses_y:
            y_arg = y
        predictions = self.estimator.predict(features_no_nan, y_arg)
        predictions = predictions.rename(self.input_target_name)
        padded = pad_with_nans(predictions, max(0, features.shape[0] - predictions.shape[0]))
        return infer_feature_types(padded)

    def score(self, X, y, objectives):
        """Evaluate model performance on current and additional objectives.

        Arguments:
            X (ww.DataTable, pd.DataFrame or np.ndarray): Data of shape [n_samples, n_features]
            y (pd.Series, ww.DataColumn): True labels of length [n_samples]
            objectives (list): Non-empty list of objectives to score on

        Returns:
            dict: Ordered dictionary of objective scores
        """
        # Only converting X for the call to _score_all_objectives
        if X is None:
            X = pd.DataFrame()
        X = infer_feature_types(X)
        y = infer_feature_types(y)

        y_predicted = self.predict(X, y)

        y_shifted = y.shift(-self.gap)
        objectives = self.create_objectives(objectives)
        y_shifted, y_predicted = drop_rows_with_nans(y_shifted, y_predicted)
        return self._score_all_objectives(X, y_shifted,
                                          y_predicted,
                                          y_pred_proba=None,
                                          objectives=objectives)



import numpy as np
from sklearn.model_selection import TimeSeriesSplit as SkTimeSeriesSplit
from sklearn.model_selection._split import BaseCrossValidator


class TimeSeriesSplit(BaseCrossValidator):
    """Rolling Origin Cross Validation for time series problems."""

    def __init__(self, max_delay=0, gap=0, n_splits=3):
        """Create a TimeSeriesSplit.

        This class uses max_delay and gap values to take into account that bciavm time series pipelines perform
        some feature and target engineering, e.g delaying input features and shifting the target variable by the
        desired amount. If the data that will be split already has all the features and appropriate target values, and
        then set max_delay and gap to 0.

        Arguments:
            max_delay (int): Max delay value for feature engineering. Time series pipelines create delayed features
                from existing features. This process will introduce NaNs into the first max_delay number of rows. The
                splitter uses the last max_delay number of rows from the previous split as the first max_delay number
                of rows of the current split to avoid "throwing out" more data than in necessary.
            gap (int): Gap used in time series problem. Time series pipelines shift the target variable by gap rows
                since we are interested in
            n_splits (int): number of data splits to make.
            """
        self.max_delay = max_delay
        self.gap = gap
        self.n_splits = n_splits
        self._splitter = SkTimeSeriesSplit(n_splits=n_splits)

    def get_n_splits(self, X=None, y=None, groups=None):
        """Get the number of data splits."""
        return self._splitter.n_splits

    @staticmethod
    def _check_if_empty(data):
        return data is None or data.empty

    def split(self, X, y=None, groups=None):
        """Get the time series splits.

        X and y are assumed to be sorted in ascending time order.
        This method can handle passing in empty or None X and y data but note that X and y cannot be None or empty
        at the same time.

        Arguments:
            X (pd.DataFrame, None): Features to split.
            y (pd.DataFrame, None): Target variable to split.
            groups: Ignored but kept for compatibility with sklearn api.

        Returns:
            Iterator of (train, test) indices tuples.
        """
        # Sklearn splitters always assume a valid X is passed but we need to support the
        # TimeSeriesPipeline convention of being able to pass in empty X dataframes
        # We'll do this by passing X=y if X is empty
        if self._check_if_empty(X) and self._check_if_empty(y):
            raise ValueError("Both X and y cannot be None or empty in TimeSeriesSplit.split")
        elif self._check_if_empty(X) and not self._check_if_empty(y):
            split_kwargs = dict(X=y, groups=groups)
            max_index = y.shape[0]
        else:
            split_kwargs = dict(X=X, y=y, groups=groups)
            max_index = X.shape[0]

        split_size = (max_index // self.n_splits)
        if split_size < self.gap + self.max_delay:
            raise ValueError(f"Since the data has {max_index} observations and n_splits={self.n_splits}, "
                             f"the smallest split would have {split_size} observations. "
                             f"Since {self.gap + self.max_delay} (gap + max_delay)  > {split_size}, "
                             "then at least one of the splits would be empty by the time it reaches the pipeline. "
                             "Please use a smaller number of splits or collect more data.")

        for train, test in self._splitter.split(**split_kwargs):
            last_train = train[-1]
            first_test, last_test = test[0], test[-1]

            # Make sure the splits don't extend past our data
            max_test_index = min(last_test + 1 + self.gap, max_index)
            max_train_index = min(last_train + 1 + self.gap, max_index)

            new_train = np.arange(0, max_train_index)
            new_test = np.arange(first_test - self.max_delay, max_test_index)
            yield new_train, new_test



from enum import Enum

from bciavm.utils import classproperty


class ProblemTypes(Enum):
    """Enum defining the supported types of machine learning problems."""
    BINARY = 'binary'
    """Binary classification problem."""
    MULTICLASS = 'multiclass'
    """Multiclass classification problem."""
    REGRESSION = 'regression'
    """Regression problem."""
    TIME_SERIES_REGRESSION = 'time series regression'
    """Time series regression problem."""
    TIME_SERIES_BINARY = 'time series binary'
    """Time series binary classification problem."""
    TIME_SERIES_MULTICLASS = 'time series multiclass'
    """Time series multiclass classification problem."""

    def __str__(self):
        problem_type_dict = {ProblemTypes.BINARY.name: "binary",
                             ProblemTypes.MULTICLASS.name: "multiclass",
                             ProblemTypes.REGRESSION.name: "regression",
                             ProblemTypes.TIME_SERIES_REGRESSION.name: "time series regression",
                             ProblemTypes.TIME_SERIES_BINARY.name: "time series binary",
                             ProblemTypes.TIME_SERIES_MULTICLASS.name: "time series multiclass"}
        return problem_type_dict[self.name]

    @classproperty
    def _all_values(cls):
        return {pt.value.upper(): pt for pt in cls.all_problem_types}

    @classproperty
    def all_problem_types(cls):
        """Get a list of all defined problem types.

        Returns:
            list(ProblemTypes): list
        """
        return list(cls)



from .utils import (
    load_data,
    split_data,
    number_of_features,
    target_distribution,
    drop_nan_target_rows
)
from .data_splitters import (
    TrainingValidationSplit,
    TimeSeriesSplit,
    BalancedClassificationDataCVSplit,
    BalancedClassificationDataTVSplit,
    KMeansSMOTETVSplit,
    KMeansSMOTECVSplit,
    SMOTETomekTVSplit,
    SMOTETomekCVSplit,
    RandomUnderSamplerCVSplit,
    RandomUnderSamplerTVSplit,
    SMOTENCCVSplit,
    SMOTENCTVSplit
)




from .objective_base import ObjectiveBase

from bciavm import objectives
from bciavm.exceptions import ObjectiveCreationError, ObjectiveNotFoundError
from bciavm.problem_types import handle_problem_types
from bciavm.utils.gen_utils import _get_subclasses


def get_non_core_objectives():
    """Get non-core objective classes.

    Non-core objectives are objectives that are domain-specific. Users typically need to configure these objectives
    before using them in AutoMLSearch.

    Returns:
        List of ObjectiveBase classes
    """
    return [objectives.CostBenefitMatrix, objectives.FraudCost, objectives.LeadScoring,
            objectives.Recall, objectives.RecallMacro, objectives.RecallMicro,
            objectives.RecallWeighted, objectives.MAPE, objectives.MeanSquaredLogError,
            objectives.RootMeanSquaredLogError]


def _all_objectives_dict():
    all_objectives = _get_subclasses(ObjectiveBase)
    objectives_dict = {}
    for objective in all_objectives:
        if 'bciavm.objectives' not in objective.__module__:
            continue
        objectives_dict[objective.name.lower()] = objective
    return objectives_dict


def get_all_objective_names():
    """Get a list of the names of all objectives.

    Returns:
        list (str): Objective names
    """
    all_objectives_dict = _all_objectives_dict()
    return list(all_objectives_dict.keys())


def get_core_objective_names():
    """Get a list of all valid core objectives.

    Returns:
        list[str]: Objective names.
    """
    all_objectives = _all_objectives_dict()
    non_core = get_non_core_objectives()
    return [name for name, class_name in all_objectives.items() if class_name not in non_core]


def get_objective(objective, return_instance=False, **kwargs):
    """Returns the Objective class corresponding to a given objective name.

    Arguments:
        objective (str or ObjectiveBase): Name or instance of the objective class.
        return_instance (bool): Whether to return an instance of the objective. This only applies if objective
            is of type str. Note that the instance will be initialized with default arguments.
        kwargs (Any): Any keyword arguments to pass into the objective. Only used when return_instance=True.

    Returns:
        ObjectiveBase if the parameter objective is of type ObjectiveBase. If objective is instead a valid
        objective name, function will return the class corresponding to that name. If return_instance is True,
        an instance of that objective will be returned.
    """
    if objective is None:
        raise TypeError("Objective parameter cannot be NoneType")
    if isinstance(objective, ObjectiveBase):
        return objective
    all_objectives_dict = _all_objectives_dict()
    if not isinstance(objective, str):
        raise TypeError("If parameter objective is not a string, it must be an instance of ObjectiveBase!")
    if objective.lower() not in all_objectives_dict:
        raise ObjectiveNotFoundError(f"{objective} is not a valid Objective! "
                                     "Use bciavm.objectives.get_all_objective_names()"
                                     "to get a list of all valid objective names. ")

    objective_class = all_objectives_dict[objective.lower()]

    if return_instance:
        try:
            return objective_class(**kwargs)
        except TypeError as e:
            raise ObjectiveCreationError(f"In get_objective, cannot pass in return_instance=True for {objective} because {str(e)}")

    return objective_class


def get_core_objectives(problem_type):
    """Returns all core objective instances associated with the given problem type.

    Core objectives are designed to work out-of-the-box for any dataset.

    Arguments:
        problem_type (str/ProblemTypes): Type of problem

    Returns:
        List of ObjectiveBase instances
    """
    problem_type = handle_problem_types(problem_type)
    all_objectives_dict = _all_objectives_dict()
    objectives = [obj() for obj in all_objectives_dict.values() if obj.is_defined_for_problem_type(problem_type) and obj not in get_non_core_objectives()]
    return objectives



from bciavm.preprocessing.data_splitters.base_splitters import (
    BaseSamplingSplitter
)
from bciavm.utils import import_or_raise


class SMOTETomekTVSplit(BaseSamplingSplitter):
    """Splits the data into training and validation sets and uses SMOTE + Tomek links balance the training data.
       Keeps the validation data the same. Works only on continuous, numeric data."""

    def __init__(self, sampling_strategy='auto', test_size=None, n_jobs=-1, random_seed=0):
        error_msg = "imbalanced-learn is not installed. Please install using 'pip install imbalanced-learn'"
        im = import_or_raise("imblearn.combine", error_msg=error_msg)
        self.sampler = im.SMOTETomek(sampling_strategy=sampling_strategy, n_jobs=n_jobs, random_state=random_seed)
        super().__init__(sampler=self.sampler, test_size=test_size, split_type="TV", random_seed=random_seed)


class SMOTETomekCVSplit(BaseSamplingSplitter):
    """Splits the data into KFold cross validation sets and uses SMOTE + Tomek links to balance the training data.
       Keeps the validation data the same. Works only on continuous, numeric data."""

    def __init__(self, sampling_strategy='auto', n_splits=3, shuffle=True, n_jobs=-1, random_seed=0):
        error_msg = "imbalanced-learn is not installed. Please install using 'pip install imbalanced-learn'"
        im = import_or_raise("imblearn.combine", error_msg=error_msg)
        self.sampler = im.SMOTETomek(sampling_strategy=sampling_strategy, n_jobs=n_jobs, random_state=random_seed)
        super().__init__(sampler=self.sampler, n_splits=n_splits, shuffle=shuffle, split_type="CV", random_seed=random_seed)



from sklearn.neighbors import KNeighborsRegressor as SKKNeighborsRegressor
from bciavm.model_family import ModelFamily
from bciavm.pipelines.components.estimators import Estimator
from bciavm.problem_types import ProblemTypes


class KNeighborsRegressor(Estimator):
    """KNeighborsRegressor."""
    name = "K Nearest Neighbors Regressor"
    hyperparameter_ranges = {}
    model_family = ModelFamily.LINEAR_MODEL
    supported_problem_types = [ProblemTypes.REGRESSION, ProblemTypes.TIME_SERIES_REGRESSION]

    def __init__(self,
                 n_neighbors=5,
                 weights='distance',
                 algorithm='auto',
                 p=2,
                 leaf_size=20,
                 metric='minkowski',
                 n_jobs=4,
                 random_seed=0,
                 **kwargs):

        parameters = {
            'n_neighbors': n_neighbors,
            'weights': weights,
            'algorithm': algorithm,
            'leaf_size': leaf_size,
            'p':p,
            'metric':metric,
            'n_jobs':n_jobs
        }

        parameters.update(kwargs)
        neighbors_regressor = SKKNeighborsRegressor(**parameters)
        super().__init__(parameters=parameters,
                         component_obj=neighbors_regressor,
                         random_seed=random_seed)

    def fit(self, X, y=None):
        X, y = super()._manage_woodwork(X, y)
        X = self._ensure_types(X)
        self.input_feature_names = list(X.columns)
        self._component_obj.fit(X, y)
        return self

    def _ensure_types(self, X):
        for col in X.columns:
            try: X[col] = X[col].astype(float)
            except: X = X.drop(col, axis=1)
            if col in ['unit_indx','_c0']:
                try: X = X.drop(col, axis=1)
                except: pass
        return X

    def predict(self, X):
        X = self._ensure_types(X)
        X = X[self.input_feature_names]
        return super().predict(X)

    @property
    def feature_importance(self):
        return self._component_obj.coef_



from .objective_base import ObjectiveBase

from bciavm.problem_types import ProblemTypes


class MulticlassClassificationObjective(ObjectiveBase):
    """Base class for all multiclass classification objectives."""

    problem_types = [ProblemTypes.MULTICLASS, ProblemTypes.TIME_SERIES_MULTICLASS]



from bciavm.preprocessing.data_splitters.base_splitters import (
    BaseSamplingSplitter
)
from bciavm.utils import import_or_raise


def _allowed_categorical(categorical_features):
    if categorical_features is None:
        return False
    elif not isinstance(categorical_features, list):
        return False
    elif len(categorical_features) == 0:
        return False
    elif (all(categorical_features) and all([isinstance(c, bool) for c in categorical_features])):
        return False
    return True


class SMOTENCTVSplit(BaseSamplingSplitter):
    """Splits the data into training and validation sets and uses SMOTENC to balance the training data.
       Keeps the validation data the same. Works on numeric and categorical data, but categorical data must be numerical"""

    def __init__(self, categorical_features=None, sampling_strategy='auto', test_size=None, n_jobs=-1, random_seed=0):
        error_msg = "imbalanced-learn is not installed. Please install using 'pip install imbalanced-learn'"
        im = import_or_raise("imblearn.over_sampling", error_msg=error_msg)
        if not _allowed_categorical(categorical_features):
            raise ValueError(f"Categorical feature array must be a list with values and must not all be True, received {categorical_features}")
        self.categorical_features = categorical_features
        self.sampler = im.SMOTENC(categorical_features=self.categorical_features,
                                  sampling_strategy=sampling_strategy,
                                  n_jobs=n_jobs, random_state=random_seed)
        super().__init__(sampler=self.sampler, test_size=test_size, split_type="TV", random_seed=random_seed)


class SMOTENCCVSplit(BaseSamplingSplitter):
    """Splits the data into KFold cross validation sets and uses SMOTENC to balance the training data.
       Keeps the validation data the same. Works on numeric and categorical data, but categorical data must be numerical"""

    def __init__(self, categorical_features=None, sampling_strategy='auto', n_splits=3, shuffle=True, n_jobs=-1, random_seed=0):
        error_msg = "imbalanced-learn is not installed. Please install using 'pip install imbalanced-learn'"
        im = import_or_raise("imblearn.over_sampling", error_msg=error_msg)
        if not _allowed_categorical(categorical_features):
            raise ValueError(f"Categorical feature array must be a list with values and must not all be True, received {categorical_features}")
        self.categorical_features = categorical_features
        self.sampler = im.SMOTENC(categorical_features=self.categorical_features,
                                  sampling_strategy=sampling_strategy,
                                  n_jobs=n_jobs, random_state=random_seed)
        super().__init__(sampler=self.sampler, n_splits=n_splits, shuffle=shuffle, split_type="CV", random_seed=random_seed)



import pandas as pd
import woodwork as ww
from skopt.space import Integer

from bciavm.pipelines.components.transformers.transformer import Transformer
from bciavm.utils import import_or_raise, infer_feature_types


class PolynomialDetrender(Transformer):
    """Removes trends from time series by fitting a polynomial to the data."""
    name = "Polynomial Detrender"

    hyperparameter_ranges = {"degree": Integer(1, 3)}

    def __init__(self, degree=1, random_seed=0, **kwargs):
        """Initialize the PolynomialDetrender.

        Arguments:
            degree (int): Degree for the polynomial. If 1, linear model is fit to the data.
                If 2, quadratic model is fit, etc. Default of 1.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        """
        if not isinstance(degree, int):
            if isinstance(degree, float) and degree.is_integer():
                degree = int(degree)
            else:
                raise TypeError(f"Parameter Degree must be an integer!: Received {type(degree).__name__}")

        params = {"degree": degree}
        params.update(kwargs)
        error_msg = "sktime is not installed. Please install using 'pip install sktime'"

        trend = import_or_raise("sktime.forecasting.trend", error_msg=error_msg)
        detrend = import_or_raise("sktime.transformations.series.detrend", error_msg=error_msg)

        detrender = detrend.Detrender(trend.PolynomialTrendForecaster(degree=degree))

        super().__init__(parameters=params,
                         component_obj=detrender,
                         random_seed=random_seed)

    def fit(self, X, y=None):
        """Fits the PolynomialDetrender.

        Arguments:
            X (ww.DataTable, pd.DataFrame, optional): Ignored.
            y (ww.DataColumn, pd.Series): Target variable to detrend.

        Returns:
            self
        """
        if y is None:
            raise ValueError("y cannot be None for PolynomialDetrender!")
        y_dt = infer_feature_types(y)
        self._component_obj.fit(y_dt)
        return self

    def transform(self, X, y=None):
        """Removes fitted trend from target variable.

        Arguments:
            X (ww.DataTable, pd.DataFrame, optional): Ignored.
            y (ww.DataColumn, pd.Series): Target variable to detrend.

        Returns:
            tuple of ww.DataTable, ww.DataColumn: The input features are returned without modification. The target
                variable y is detrended
        """
        if y is None:
            return X, y
        y_dt = infer_feature_types(y)
        y_t = self._component_obj.transform(y_dt)
        y_t = infer_feature_types(pd.Series(y_t, index=y_dt.index))
        return X, y_t

    def fit_transform(self, X, y=None):
        """Removes fitted trend from target variable.

        Arguments:
            X (ww.DataTable, pd.DataFrame, optional): Ignored.
            y (ww.DataColumn, pd.Series): Target variable to detrend.

        Returns:
            tuple of ww.DataTable, ww.DataColumn: The first element are the input features returned without modification.
                The second element is the target variable y with the fitted trend removed.
        """
        return self.fit(X, y).transform(X, y)

    def inverse_transform(self, X, y):
        """Adds back fitted trend to target variable.

        Arguments:
            X (ww.DataTable, pd.DataFrame, optional): Ignored.
            y (ww.DataColumn, pd.Series): Target variable.

        Returns:
            tuple of ww.DataTable, ww.DataColumn: The first element are the input features returned without modification.
                The second element is the target variable y with the trend added back.
        """
        if y is None:
            raise ValueError("y cannot be None for PolynomialDetrender!")
        y_dt = infer_feature_types(y)
        y_t = self._component_obj.inverse_transform(y_dt)
        y_t = infer_feature_types(pd.Series(y_t, index=y_dt.index))
        return X, y_t



import pandas as pd

from bciavm.pipelines.components.transformers import Transformer
from bciavm.pipelines.components.transformers.imputers import SimpleImputer
from bciavm import Boolean, Categorical, Double, Integer
from bciavm.utils import infer_feature_types

numeric_dtypes = [Integer, Double, Boolean]
categorical_dtypes = [Categorical]


class Imputer(Transformer):
    """Imputes missing data according to a specified imputation strategy."""
    name = "Imputer"
    hyperparameter_ranges = {
        "categorical_impute_strategy": ["most_frequent"],
        "numeric_impute_strategy": ["mean", "median", "most_frequent"]
    }
    _valid_categorical_impute_strategies = set(["most_frequent", "constant"])
    _valid_numeric_impute_strategies = set(["mean", "median", "most_frequent", "constant"])

    def __init__(self, categorical_impute_strategy="most_frequent",
                 categorical_fill_value=None,
                 numeric_impute_strategy="mean",
                 numeric_fill_value=None,
                 random_state=0, **kwargs):
        """Initalizes an transformer that imputes missing data according to the specified imputation strategy."

        Arguments:
            categorical_impute_strategy (string): Impute strategy to use for string, object, boolean, categorical dtypes. Valid values include "most_frequent" and "constant".
            numeric_impute_strategy (string): Impute strategy to use for numeric columns. Valid values include "mean", "median", "most_frequent", and "constant".
            categorical_fill_value (string): When categorical_impute_strategy == "constant", fill_value is used to replace missing data. The default value of None will fill with the string "missing_value".
            numeric_fill_value (int, float): When numeric_impute_strategy == "constant", fill_value is used to replace missing data. The default value of None will fill with 0.
        """
        if categorical_impute_strategy not in self._valid_categorical_impute_strategies:
            raise ValueError(f"{categorical_impute_strategy} is an invalid parameter. Valid categorical impute strategies are {', '.join(self._valid_numeric_impute_strategies)}")
        elif numeric_impute_strategy not in self._valid_numeric_impute_strategies:
            raise ValueError(f"{numeric_impute_strategy} is an invalid parameter. Valid impute strategies are {', '.join(self._valid_numeric_impute_strategies)}")

        parameters = {"categorical_impute_strategy": categorical_impute_strategy,
                      "numeric_impute_strategy": numeric_impute_strategy,
                      "categorical_fill_value": categorical_fill_value,
                      "numeric_fill_value": numeric_fill_value}
        parameters.update(kwargs)
        self._categorical_imputer = SimpleImputer(impute_strategy=categorical_impute_strategy,
                                                  fill_value=categorical_fill_value,
                                                  **kwargs)
        self._numeric_imputer = SimpleImputer(impute_strategy=numeric_impute_strategy,
                                              fill_value=numeric_fill_value,
                                              **kwargs)
        self._all_null_cols = None
        self._numeric_cols = None
        self._categorical_cols = None
        super().__init__(parameters=parameters,
                         component_obj=None,
                         random_state=random_state)

    def fit(self, X, y=None):
        """Fits imputer to data. 'None' values are converted to np.nan before imputation and are
            treated as the same.

        Arguments:
            X (pd.DataFrame or np.array): The input training data of shape [n_samples, n_features]
            y (pd.Series, optional): The target training data of length [n_samples]

        Returns:
            self
        """
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)

        self._all_null_cols = set(X.columns) - set(X.dropna(axis=1, how='all').columns)
        X_copy = X.copy()
        X_null_dropped = X_copy.drop(self._all_null_cols, axis=1, errors='ignore')

        numerics = [ 'int16', 'int32', 'int64', 'float16', 'float32', 'float64']

        X_numerics = X_null_dropped.select_dtypes(include=numerics)
        if len(X_numerics.columns) > 0:
            self._numeric_imputer.fit(X_numerics, y)
            self._numeric_cols = X_numerics.columns

        categorical_dtypes = [ 'category', 'object', 'boolean' ]
        X_categorical = X_null_dropped.select_dtypes(include=categorical_dtypes)
        if len(X_categorical.columns) > 0:
            self._categorical_imputer.fit(X_categorical, y)
            self._categorical_cols = X_categorical.columns
        return self

    def transform(self, X, y=None):
        """Transforms data X by imputing missing values. 'None' values are converted to np.nan before imputation and are
            treated as the same.

        Arguments:
            X (pd.DataFrame): Data to transform
            y (pd.Series, optional): Ignored.

        Returns:
            pd.DataFrame: Transformed X
        """
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)

        X_null_dropped = X.copy()
        X_null_dropped.drop(self._all_null_cols, inplace=True, axis=1, errors='ignore')
        X_null_dropped.reset_index(inplace=True, drop=True)
        if X_null_dropped.empty:
            return X_null_dropped

        if 'unit_indx' not in X_null_dropped.columns:
            X_null_dropped = X_null_dropped.reset_index().rename({'index':'unit_indx'},axis=1)

        if self._numeric_cols is not None and len(self._numeric_cols) > 0:
            X_numeric = X_null_dropped[self._numeric_cols]
            X_null_dropped[X_numeric.columns] = self._numeric_imputer.transform(X_numeric)

        if self._categorical_cols is not None and len(self._categorical_cols) > 0:
            X_categorical = X_null_dropped[self._categorical_cols]
            X_null_dropped[X_categorical.columns] = self._categorical_imputer.transform(X_categorical)
        X_null_dropped = infer_feature_types(X_null_dropped)
        return X_null_dropped



import pandas as pd
from sklearn.decomposition import TruncatedSVD
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.pipeline import make_pipeline

from bciavm.pipelines.components.transformers.preprocessing import (
    TextTransformer
)


class LSA(TextTransformer):
    """Transformer to calculate the Latent Semantic Analysis Values of text input"""
    name = "LSA Transformer"
    hyperparameter_ranges = {}

    def __init__(self, text_columns=None, random_state=0, **kwargs):
        """Creates a transformer to perform TF-IDF transformation and Singular Value Decomposition for text columns.

        Arguments:
            text_columns (list): list of feature names which should be treated as text features.
            random_state (int, np.random.RandomState): Seed for the random number generator.
        """
        self._lsa_pipeline = make_pipeline(TfidfVectorizer(), TruncatedSVD(random_state=random_state))
        super().__init__(text_columns=text_columns,
                         random_state=random_state,
                         **kwargs)

    def fit(self, X, y=None):
        if len(self._all_text_columns) == 0:
            return self
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        text_columns = self._get_text_columns(X)
        corpus = X[text_columns].values.flatten()
        # we assume non-str values will have been filtered out prior to calling LSA.fit. this is a safeguard.
        corpus = corpus.astype(str)
        self._lsa_pipeline.fit(corpus)
        return self

    def transform(self, X, y=None):
        """Transforms data X by applying the LSA pipeline.

        Arguments:
            X (pd.DataFrame): Data to transform
            y (pd.Series, optional): Ignored.

        Returns:
            pd.DataFrame: Transformed X. The original column is removed and replaced with two columns of the
                          format `LSA(original_column_name)[feature_number]`, where `feature_number` is 0 or 1.
        """
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        if len(self._all_text_columns) == 0:
            return X

        X_t = X.copy()
        text_columns = self._get_text_columns(X)
        for col in text_columns:
            transformed = self._lsa_pipeline.transform(X[col].astype(str).fillna('None'))

            X_t['LSA({})[0]'.format(col)] = pd.Series(transformed[:, 0])
            X_t['LSA({})[1]'.format(col)] = pd.Series(transformed[:, 1])
        X_t = X_t.drop(columns=text_columns)
        return X_t



import pandas as pd
from woodwork.logical_types import Categorical

from ..transformer import Transformer

from bciavm.pipelines.components.transformers.encoders.onehot_encoder import (
    OneHotEncoderMeta
)
from bciavm.utils import (
    _retain_custom_types_and_initalize_woodwork,
    import_or_raise,
    infer_feature_types
)


class TargetEncoder(Transformer, metaclass=OneHotEncoderMeta):
    """Target encoder to encode categorical data"""
    name = 'Target Encoder'
    hyperparameter_ranges = {}

    def __init__(self,
                 cols=None,
                 smoothing=1.0,
                 handle_unknown='value',
                 handle_missing='value',
                 random_seed=0,
                 **kwargs):
        """Initializes a transformer that encodes categorical features into target encodings.

        Arguments:
            cols (list): Columns to encode. If None, all string columns will be encoded, otherwise only the columns provided will be encoded.
                Defaults to None
            smoothing (float): The smoothing factor to apply. The larger this value is, the more influence the expected target value has
                on the resulting target encodings. Must be strictly larger than 0. Defaults to 1.0
            handle_unknown (string): Determines how to handle unknown categories for a feature encountered. Options are 'value', 'error', nd 'return_nan'.
                Defaults to 'value', which replaces with the target mean
            handle_missing (string): Determines how to handle missing values encountered during `fit` or `transform`. Options are 'value', 'error', and 'return_nan'.
                Defaults to 'value', which replaces with the target mean
            random_seed (int): Seed for the random number generator. Defaults to 0.
            """

        parameters = {"cols": cols,
                      "smoothing": smoothing,
                      "handle_unknown": handle_unknown,
                      "handle_missing": handle_missing}
        parameters.update(kwargs)

        unknown_and_missing_input_options = ['error', 'return_nan', 'value']
        if handle_unknown not in unknown_and_missing_input_options:
            raise ValueError("Invalid input '{}' for handle_unknown".format(handle_unknown))
        if handle_missing not in unknown_and_missing_input_options:
            raise ValueError("Invalid input '{}' for handle_missing".format(handle_missing))
        if smoothing <= 0:
            raise ValueError("Smoothing value needs to be strictly larger than 0. {} provided".format(smoothing))

        category_encode = import_or_raise('category_encoders', error_msg='category_encoders not installed. Please install using `pip install category_encoders`')
        super().__init__(parameters=parameters,
                         component_obj=category_encode.target_encoder.TargetEncoder(**parameters),
                         random_seed=random_seed)

    def fit(self, X, y):
        return super().fit(X, y)

    def transform(self, X, y=None):
        X_ww = infer_feature_types(X)
        if y is not None:
            y = infer_feature_types(y)
        X_t = self._component_obj.transform(X, y)
        X_t_df = pd.DataFrame(X_t, columns=X_ww.columns, index=X_ww.index)
        return _retain_custom_types_and_initalize_woodwork(X_ww.ww.logical_types, X_t_df, ltypes_to_ignore=[Categorical])

    def fit_transform(self, X, y):
        return self.fit(X, y).transform(X, y)

    def get_feature_names(self):
        """Return feature names for the input features after fitting.

        Returns:
            np.array: The feature names after encoding
        """
        return self._component_obj.get_feature_names()

    def _get_feature_provenance(self):
        return {col: col for col in self.get_feature_names()}



import json

from woodwork import logical_types

from .regression_pipeline import RegressionPipeline

from .time_series_regression_pipeline import TimeSeriesRegressionPipeline

from bciavm import Boolean, Integer
from bciavm.model_family import ModelFamily
from bciavm.pipelines import PipelineBase
from bciavm.pipelines.components import (  # noqa: F401
    ComponentBase,
    DateTimeFeaturizer,
    DelayedFeatureTransformer,
    DropColumns,
    DropNullColumns,
    Estimator,
    Imputer,
    OneHotEncoder,
    StackedEnsembleRegressor,
    StandardScaler,
    TextFeaturizer
)
from bciavm.pipelines.components.utils import all_components, get_estimators
from bciavm.problem_types import (
    ProblemTypes,
    handle_problem_types,
    is_time_series
)
from bciavm.utils import get_logger, infer_feature_types

logger = get_logger(__file__)


def _get_preprocessing_components(X, y, problem_type, estimator_class):
    """Given input data, target data and an estimator class, construct a recommended preprocessing chain to be combined with the estimator and trained on the provided data.

    Arguments:
        X (ww.DataTable): The input data of shape [n_samples, n_features]
        y (ww.DataColumn): The target data of length [n_samples]
        problem_type (ProblemTypes or str): Problem type
        estimator_class (class): A class which subclasses Estimator estimator for pipeline

    Returns:
        list[Transformer]: A list of applicable preprocessing components to use with the estimator
    """

    pp_components = []
    all_null_cols = X.columns[X.isnull().all()]
    if len(all_null_cols) > 0:
        pp_components.append(DropNullColumns)
    input_logical_types = set(X.ww.logical_types.values())
    types_imputer_handles = {Boolean, logical_types.Categorical, logical_types.Double, Integer}
    if len(input_logical_types.intersection(types_imputer_handles)) > 0:
        pp_components.append(Imputer)

    text_columns = list(X.ww.select('NaturalLanguage').columns)
    if len(text_columns) > 0:
        pp_components.append(TextFeaturizer)

    datetime_cols = X.ww.select(["Datetime"])
    add_datetime_featurizer = len(datetime_cols.columns) > 0
    if add_datetime_featurizer:
        pp_components.append(DateTimeFeaturizer)

    if is_time_series(problem_type):
        pp_components.append(DelayedFeatureTransformer)

    categorical_cols = X.ww.select('category')
    if len(categorical_cols.columns) > 0:
        pp_components.append(OneHotEncoder)

    if estimator_class.model_family == ModelFamily.LINEAR_MODEL:
        pp_components.append(StandardScaler)
    return pp_components


def _get_pipeline_base_class(problem_type):
    """Returns pipeline base class for problem_type"""
    if problem_type == ProblemTypes.REGRESSION:
        return RegressionPipeline
    elif problem_type == ProblemTypes.TIME_SERIES_REGRESSION:
        return TimeSeriesRegressionPipeline



def make_pipeline(X, y, estimator, problem_type, custom_hyperparameters=None):
    """Given input data, target data, an estimator class and the problem type,
        generates a pipeline class with a preprocessing chain which was recommended based on the inputs.
        The pipeline will be a subclass of the appropriate pipeline base class for the specified problem_type.

   Arguments:
        X (pd.DataFrame, ww.DataTable): The input data of shape [n_samples, n_features]
        y (pd.Series, ww.DataColumn): The target data of length [n_samples]
        estimator (Estimator): Estimator for pipeline
        problem_type (ProblemTypes or str): Problem type for pipeline to generate
        custom_hyperparameters (dictionary): Dictionary of custom hyperparameters,
            with component name as key and dictionary of parameters as the value

    Returns:
        class: PipelineBase subclass with dynamically generated preprocessing components and specified estimator

    """
    X = infer_feature_types(X)
    y = infer_feature_types(y)

    problem_type = handle_problem_types(problem_type)
    if estimator not in get_estimators(problem_type):
        raise ValueError(f"{estimator.name} is not a valid estimator for problem type")
    preprocessing_components = _get_preprocessing_components(X, y, problem_type, estimator)
    complete_component_graph = preprocessing_components + [estimator]

    if custom_hyperparameters and not isinstance(custom_hyperparameters, dict):
        raise ValueError(f"if custom_hyperparameters provided, must be dictionary. Received {type(custom_hyperparameters)}")

    hyperparameters = custom_hyperparameters
    base_class = _get_pipeline_base_class(problem_type)

    class GeneratedPipeline(base_class):
        custom_name = f"{estimator.name} w/ {' + '.join([component.name for component in preprocessing_components])}"
        component_graph = complete_component_graph
        custom_hyperparameters = hyperparameters

    return GeneratedPipeline


def make_pipeline_from_components(component_instances, problem_type, custom_name=None, random_seed=0):
    """Given a list of component instances and the problem type, an pipeline instance is generated with the component instances.
    The pipeline will be a subclass of the appropriate pipeline base class for the specified problem_type. The pipeline will be
    untrained, even if the input components are already trained. A custom name for the pipeline can optionally be specified;
    otherwise the default pipeline name will be 'Templated Pipeline'.

   Arguments:
        component_instances (list): a list of all of the components to include in the pipeline
        problem_type (str or ProblemTypes): problem type for the pipeline to generate
        custom_name (string): a name for the new pipeline
        random_seed (int): Random seed used to intialize the pipeline.

    Returns:
        Pipeline instance with component instances and specified estimator created from given random state.

    Example:
        >>> components = [Imputer(), StandardScaler(), RandomForestClassifier()]
        >>> pipeline = make_pipeline_from_components(components, problem_type="binary")
        >>> pipeline.describe()

    """
    for i, component in enumerate(component_instances):
        if not isinstance(component, ComponentBase):
            raise TypeError("Every element of `component_instances` must be an instance of ComponentBase")
        if i == len(component_instances) - 1 and not isinstance(component, Estimator):
            raise ValueError("Pipeline needs to have an estimator at the last position of the component list")

    if custom_name and not isinstance(custom_name, str):
        raise TypeError("Custom pipeline name must be a string")
    pipeline_name = custom_name
    problem_type = handle_problem_types(problem_type)

    class TemplatedPipeline(_get_pipeline_base_class(problem_type)):
        custom_name = pipeline_name
        component_graph = [c.__class__ for c in component_instances]
    return TemplatedPipeline({c.name: c.parameters for c in component_instances}, random_seed=random_seed)


def generate_pipeline_code(element):
    """Creates and returns a string that contains the Python imports and code required for running the bciavm pipeline.

    Arguments:
        element (pipeline instance): The instance of the pipeline to generate string Python code

    Returns:
        String representation of Python code that can be run separately in order to recreate the pipeline instance.
        Does not include code for custom component implementation.
    """
    # hold the imports needed and add code to end
    code_strings = ['import json']
    if not isinstance(element, PipelineBase):
        raise ValueError("Element must be a pipeline instance, received {}".format(type(element)))
    if isinstance(element.component_graph, dict):
        raise ValueError("Code generation for nonlinear pipelines is not supported yet")

    component_graph_string = ',\n\t\t'.join([com.__class__.__name__ if com.__class__ not in all_components() else "'{}'".format(com.name) for com in element])
    code_strings.append("from {} import {}".format(element.__class__.__bases__[0].__module__, element.__class__.__bases__[0].__name__))
    # check for other attributes associated with pipeline (ie name, custom_hyperparameters)
    pipeline_list = []
    for k, v in sorted(list(filter(lambda item: item[0][0] != '_', element.__class__.__dict__.items())), key=lambda x: x[0]):
        if k == 'component_graph':
            continue
        pipeline_list += ["{} = '{}'".format(k, v)] if isinstance(v, str) else ["{} = {}".format(k, v)]

    pipeline_string = "\t" + "\n\t".join(pipeline_list) + "\n" if len(pipeline_list) else ""

    try:
        ret = json.dumps(element.parameters, indent='\t')
    except TypeError:
        raise TypeError(f"Value {element.parameters} cannot be JSON-serialized")
    # create the base string for the pipeline
    base_string = "\nclass {0}({1}):\n" \
                  "\tcomponent_graph = [\n\t\t{2}\n\t]\n" \
                  "{3}" \
                  "\nparameters = json.loads(\"\"\"{4}\"\"\")\n" \
                  "pipeline = {0}(parameters)" \
                  .format(element.__class__.__name__,
                          element.__class__.__bases__[0].__name__,
                          component_graph_string,
                          pipeline_string,
                          ret)
    code_strings.append(base_string)
    return "\n".join(code_strings)


def _make_stacked_ensemble_pipeline(input_pipelines, problem_type, n_jobs=-1, random_seed=0):
    """
    Creates a pipeline with a stacked ensemble estimator.

    Arguments:
        input_pipelines (list(PipelineBase or subclass obj)): List of pipeline instances to use as the base estimators for the stacked ensemble.
            This must not be None or an empty list or else EnsembleMissingPipelinesError will be raised.
        problem_type (ProblemType): problem type of pipeline
        n_jobs (int or None): Integer describing level of parallelism used for pipelines.
            None and 1 are equivalent. If set to -1, all CPUs are used. For n_jobs below -1, (n_cpus + 1 + n_jobs) are used.
            Defaults to -1.

    Returns:
        Pipeline with appropriate stacked ensemble estimator.
    """
    return make_pipeline_from_components([StackedEnsembleRegressor(input_pipelines, n_jobs=n_jobs)], problem_type,
                                             custom_name="Stacked Ensemble Regression Pipeline",
                                             random_seed=random_seed)


def _make_component_list_from_actions(actions):
    """
    Creates a list of components from the input DataCheckAction list

    Arguments:
        actions (list(DataCheckAction)): List of DataCheckAction objects used to create list of components

    Returns:
        List of components used to address the input actions
    """
    components = []

    return components





from functools import wraps

from bciavm.exceptions import ComponentNotYetFittedError
from bciavm.utils.base_meta import BaseMeta


class ComponentBaseMeta(BaseMeta):
    """Metaclass that overrides creating a new component by wrapping methods with validators and setters"""

    @classmethod
    def check_for_fit(cls, method):
        """`check_for_fit` wraps a method that validates if `self._is_fitted` is `True`.
            It raises an exception if `False` and calls and returns the wrapped method if `True`.
        """
        @wraps(method)
        def _check_for_fit(self, X=None, y=None):
            klass = type(self).__name__
            if not self._is_fitted and self.needs_fitting:
                raise ComponentNotYetFittedError(f'This {klass} is not fitted yet. You must fit {klass} before calling {method.__name__}.')
            elif method.__name__ == 'inverse_transform':
                return method(self, X, y)
            elif X is None and y is None:
                return method(self)
            elif y is None:
                return method(self, X)
            else:
                return method(self, X, y)
        return _check_for_fit



import numpy as np
from scipy.optimize import minimize_scalar

from .objective_base import ObjectiveBase

from bciavm.problem_types import ProblemTypes


class BinaryClassificationObjective(ObjectiveBase):
    """Base class for all binary classification objectives."""

    problem_types = [ProblemTypes.BINARY, ProblemTypes.TIME_SERIES_BINARY]

    @property
    def can_optimize_threshold(cls):
        """Returns a boolean determining if we can optimize the binary classification objective threshold.
            This will be false for any objective that works directly with predicted probabilities, like log loss and AUC. Otherwise, it will be true."""
        return not cls.score_needs_proba

    def optimize_threshold(self, ypred_proba, y_true, X=None):
        """Learn a binary classification threshold which optimizes the current objective.

        Arguments:
            ypred_proba (ww.DataColumn, pd.Series): The classifier's predicted probabilities
            y_true (ww.DataColumn, pd.Series): The ground truth for the predictions.
            X (ww.DataTable, pd.DataFrame, optional): Any extra columns that are needed from training data.

        Returns:
            Optimal threshold for this objective
        """
        ypred_proba = self._standardize_input_type(ypred_proba)
        y_true = self._standardize_input_type(y_true)
        if X is not None:
            X = self._standardize_input_type(X)

        if not self.can_optimize_threshold:
            raise RuntimeError("Trying to optimize objective that can't be optimized!")

        def cost(threshold):
            y_predicted = self.decision_function(ypred_proba=ypred_proba, threshold=threshold, X=X)
            cost = self.objective_function(y_true, y_predicted, X=X)
            return -cost if self.greater_is_better else cost

        optimal = minimize_scalar(cost, method='Golden', options={"maxiter": 100})
        return optimal.x

    def decision_function(self, ypred_proba, threshold=0.5, X=None):
        """Apply a learned threshold to predicted probabilities to get predicted classes.

        Arguments:
            ypred_proba (ww.DataColumn, pd.Series, np.ndarray): The classifier's predicted probabilities
            threshold (float, optional): Threshold used to make a prediction. Defaults to 0.5.
            X (ww.DataTable, pd.DataFrame, optional): Any extra columns that are needed from training data.

        Returns:
            predictions
        """
        ypred_proba = self._standardize_input_type(ypred_proba)
        return ypred_proba > threshold

    def validate_inputs(self, y_true, y_predicted):
        super().validate_inputs(y_true, y_predicted)
        if len(np.unique(y_true)) > 2:
            raise ValueError("y_true contains more than two unique values")
        if len(np.unique(y_predicted)) > 2 and not self.score_needs_proba:
            raise ValueError("y_predicted contains more than two unique values")



import pandas as pd
from sklearn.decomposition import PCA as SkPCA
from skopt.space import Real

from bciavm.pipelines.components.transformers import Transformer
from bciavm.utils import (
    _retain_custom_types_and_initalize_woodwork,
    infer_feature_types,
    is_all_numeric
)


class PCA(Transformer):
    """Reduces the number of features by using Principal Component Analysis"""
    name = 'PCA Transformer'
    hyperparameter_ranges = {
        "variance": Real(0.25, 1)}

    def __init__(self, variance=0.95, n_components=None, random_seed=0, **kwargs):
        """Initalizes an transformer that reduces the number of features using PCA."

        Arguments:
            variance (float): The percentage of the original data variance that should be preserved when reducing the
                              number of features.
            n_components (int): The number of features to maintain after computing SVD. Defaults to None, but will override
                                variance variable if set.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        """
        parameters = {"variance": variance,
                      "n_components": n_components}
        parameters.update(kwargs)
        if n_components:
            pca = SkPCA(n_components=n_components, random_state=random_seed, **kwargs)
        else:
            pca = SkPCA(n_components=variance, random_state=random_seed, **kwargs)
        super().__init__(parameters=parameters,
                         component_obj=pca,
                         random_seed=random_seed)

    def fit(self, X, y=None):
        X = infer_feature_types(X)
        if not is_all_numeric(X):
            raise ValueError("PCA input must be all numeric")
        self._component_obj.fit(X)
        return self

    def transform(self, X, y=None):
        X_ww = infer_feature_types(X)
        if not is_all_numeric(X_ww):
            raise ValueError("PCA input must be all numeric")
        X_t = self._component_obj.transform(X)
        X_t = pd.DataFrame(X_t, index=X_ww.index, columns=[f"component_{i}" for i in range(X_t.shape[1])])
        return _retain_custom_types_and_initalize_woodwork(X_ww, X_t)

    def fit_transform(self, X, y=None):
        X_ww = infer_feature_types(X)
        if not is_all_numeric(X_ww):
            raise ValueError("PCA input must be all numeric")
        X_t = self._component_obj.fit_transform(X, y)
        X_t = pd.DataFrame(X_t, index=X_ww.index, columns=[f"component_{i}" for i in range(X_t.shape[1])])
        return _retain_custom_types_and_initalize_woodwork(X_ww, X_t)




import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder as SKOneHotEncoder

from bciavm.pipelines.components.transformers import Transformer

from bciavm.pipelines.components import ComponentBaseMeta
from bciavm.utils import infer_feature_types, get_random_state


class OneHotEncoderMeta(ComponentBaseMeta):
    """A version of the ComponentBaseMeta class which includes validation on an additional one-hot-encoder-specific method `categories`"""
    METHODS_TO_CHECK = ComponentBaseMeta.METHODS_TO_CHECK + ['categories', 'get_feature_names']


class OneHotEncoder(Transformer, metaclass=OneHotEncoderMeta):
    """One-hot encoder to encode non-numeric data."""
    name = 'One Hot Encoder'
    hyperparameter_ranges = {}
    random_state = get_random_state(0)
    def __init__(self,
                 top_n=100,
                 features_to_encode=['agg_cat',
                                     ],
                 categories=None,
                 drop=None,
                 handle_unknown="ignore",
                 handle_missing="error",
                 random_state=0,
                 **kwargs):
        """Initalizes an transformer that encodes categorical features in a one-hot numeric array."

        Arguments:
            top_n (int): Number of categories per column to encode. If None, all categories will be encoded.
                Otherwise, the `n` most frequent will be encoded and all others will be dropped. Defaults to 10.
            features_to_encode (list(str)): List of columns to encode. All other columns will remain untouched.
                If None, all appropriate columns will be encoded. Defaults to None.
            categories (list): A two dimensional list of categories, where `categories[i]` is a list of the categories
                for the column at index `i`. This can also be `None`, or `"auto"` if `top_n` is not None. Defaults to None.
            drop (string): Method ("first" or "if_binary") to use to drop one category per feature. Can also be
                a list specifying which method to use for each feature. Defaults to None.
            handle_unknown (string): Whether to ignore or error for unknown categories for a feature encountered
                during `fit` or `transform`. If either `top_n` or `categories` is used to limit the number of categories
                per column, this must be "ignore". Defaults to "ignore".
            handle_missing (string): Options for how to handle missing (NaN) values encountered during
                `fit` or `transform`. If this is set to "as_category" and NaN values are within the `n` most frequent,
                "nan" values will be encoded as their own column. If this is set to "error", any missing
                values encountered will raise an error. Defaults to "error".
        """
        parameters = {"top_n": top_n,
                      "features_to_encode": features_to_encode,
                      "categories": categories,
                      "drop": drop,
                      "handle_unknown": handle_unknown,
                      "handle_missing": handle_missing}
        parameters.update(kwargs)

        # Check correct inputs
        unknown_input_options = ["ignore", "error"]
        missing_input_options = ["as_category", "error"]
        if handle_unknown not in unknown_input_options:
            raise ValueError("Invalid input {} for handle_unknown".format(handle_unknown))
        if handle_missing not in missing_input_options:
            raise ValueError("Invalid input {} for handle_missing".format(handle_missing))
        if top_n is not None and categories is not None:
            raise ValueError("Cannot use categories and top_n arguments simultaneously")

        self.features_to_encode = features_to_encode
        self._encoder = None
        super().__init__(parameters=parameters,
                         component_obj=None,
                         random_state=random_state)
        self._initial_state = self.random_state.get_state()

    @staticmethod
    def _get_cat_cols(X):
        """Get names of 'object' or 'categorical' columns in the DataFrame."""
        obj_cols = []
        for idx, dtype in enumerate(X.dtypes):
            if dtype == np.object or pd.api.types.is_categorical_dtype(dtype):
                obj_cols.append(X.columns.values[idx])
        return obj_cols

    def fit(self, X, y=None):
        top_n = self.parameters['top_n']
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        X_t = X

        if self.features_to_encode is None:
            self.features_to_encode = self._get_cat_cols(X_t)
        invalid_features = [col for col in self.features_to_encode if col not in list(X.columns)]
        if len(invalid_features) > 0:
            raise ValueError("Could not find and encode {} in input data.".format(', '.join(invalid_features)))

        if self.parameters['handle_missing'] == "as_category":
            X_t[self.features_to_encode] = X_t[self.features_to_encode].replace(np.nan, "nan")
        # elif self.parameters['handle_missing'] == "error" and X.isnull().any().any():
        #     raise ValueError("Input contains NaN")

        if len(self.features_to_encode) == 0:
            categories = 'auto'

        elif self.parameters['categories'] is not None:
            categories = self.parameters['categories']
            if len(categories) != len(self.features_to_encode) or not isinstance(categories[0], list):
                raise ValueError('Categories argument must contain a list of categories for each categorical feature')

        else:
            categories = []
            for col in X_t[self.features_to_encode]:
                value_counts = X_t[col].value_counts(dropna=False).to_frame()
                if top_n is None or len(value_counts) <= top_n:
                    unique_values = value_counts.index.tolist()
                else:
                    new_random_state = np.random.RandomState()
                    new_random_state.set_state(self._initial_state)
                    value_counts = value_counts.sample(frac=1, random_state=new_random_state)
                    value_counts = value_counts.sort_values([col], ascending=False, kind='mergesort')
                    unique_values = value_counts.head(top_n).index.tolist()
                unique_values = np.sort(unique_values)
                categories.append(unique_values)

        # Create an encoder to pass off the rest of the computation to
        self._encoder = SKOneHotEncoder(categories=categories,
                                        drop=self.parameters['drop'],
                                        handle_unknown=self.parameters['handle_unknown'])
        self._encoder.fit(X_t[self.features_to_encode])
        return self

    def transform(self, X, y=None):
        """One-hot encode the input DataFrame.

        Arguments:
            X (pd.DataFrame): Dataframe of features.
            y (pd.Series): Ignored.

        Returns:
            Transformed dataframe, where each categorical feature has been encoded into numerical columns using one-hot encoding.
        """

        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)

        cat_cols = self.features_to_encode

        if self.parameters['handle_missing'] == "as_category":
            X[cat_cols] = X[cat_cols].replace(np.nan, "nan")
        # if self.parameters['handle_missing'] == "error" and X.isnull().any().any():
        #     raise ValueError("Input contains NaN")

        X_t = pd.DataFrame()
        # Add the non-categorical columns, untouched
        for col in X.columns:
            if col not in cat_cols:
                X_t = pd.concat([X_t, X[col]], axis=1)
        # The call to pd.concat above changes the type of the index so we will manually keep it the same.
        if not X_t.empty:
            X_t.index = X.index

        # Call sklearn's transform on the categorical columns
        if len(cat_cols) > 0:
            X_cat = pd.DataFrame(self._encoder.transform(X[cat_cols]).toarray(), index=X.index)
            cat_cols_str = [str(c) for c in cat_cols]
            X_cat.columns = self._encoder.get_feature_names(input_features=cat_cols_str)
            X_t = pd.concat([X_t, X_cat], axis=1)

        X_t = infer_feature_types(X_t)
        return X_t

    def categories(self, feature_name):
        """Returns a list of the unique categories to be encoded for the particular feature, in order.

        Arguments:
            feature_name (str): the name of any feature provided to one-hot encoder during fit
        Returns:
            np.array: the unique categories, in the same dtype as they were provided during fit
        """
        try:
            index = self.features_to_encode.index(feature_name)
        except Exception:
            raise ValueError(f'Feature "{feature_name}" was not provided to one-hot encoder as a training feature')
        return self._encoder.categories_[index]

    def get_feature_names(self):
        """Return feature names for the input features after fitting.

        Returns:
            np.array: The feature names after encoding, provided in the same order as input_features.
        """
        return self._encoder.get_feature_names(self.features_to_encode)



from bciavm.pipelines.components.transformers import Transformer
from bciavm.utils import infer_feature_types


class DropNullColumns(Transformer):
    """Transformer to drop features whose percentage of NaN values exceeds a specified threshold"""
    name = "Drop Null Columns Transformer"
    hyperparameter_ranges = {}

    def __init__(self, pct_null_threshold=1.0, random_seed=0, **kwargs):
        """Initalizes an transformer to drop features whose percentage of NaN values exceeds a specified threshold.

        Arguments:
            pct_null_threshold(float): The percentage of NaN values in an input feature to drop.
                Must be a value between [0, 1] inclusive. If equal to 0.0, will drop columns with any null values.
                If equal to 1.0, will drop columns with all null values. Defaults to 0.95.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        """
        if pct_null_threshold < 0 or pct_null_threshold > 1:
            raise ValueError("pct_null_threshold must be a float between 0 and 1, inclusive.")
        parameters = {"pct_null_threshold": pct_null_threshold}
        parameters.update(kwargs)

        self._cols_to_drop = None
        super().__init__(parameters=parameters,
                         component_obj=None,
                         random_seed=random_seed)

    def fit(self, X, y=None):
        pct_null_threshold = self.parameters["pct_null_threshold"]
        X_t = infer_feature_types(X)
        percent_null = X_t.isnull().mean()
        if pct_null_threshold == 0.0:
            null_cols = percent_null[percent_null > 0]
        else:
            null_cols = percent_null[percent_null >= pct_null_threshold]
        self._cols_to_drop = list(null_cols.index)
        return self

    def transform(self, X, y=None):
        """Transforms data X by dropping columns that exceed the threshold of null values.

        Arguments:
            X (ww.DataTable, pd.DataFrame): Data to transform
            y (ww.DataColumn, pd.Series, optional): Ignored.

        Returns:
            ww.DataTable: Transformed X
        """
        X_t = infer_feature_types(X)
        if len(self._cols_to_drop) == 0:
            return X_t
        return X_t.ww.drop(self._cols_to_drop)



from bciavm.pipelines import (
    TimeSeriesRegressionPipeline
)


class TimeSeriesBaselineRegressionPipeline(TimeSeriesRegressionPipeline):
    """Baseline Pipeline for time series regression problems."""
    _name = "Time Series Baseline Regression Pipeline"
    component_graph = ["Time Series Baseline Estimator"]





import copy
from abc import ABC, abstractmethod

import cloudpickle

from bciavm.exceptions import MethodPropertyNotFoundError
from bciavm.pipelines.components.component_base_meta import ComponentBaseMeta
from bciavm.utils import (
    classproperty,
    get_logger,
    infer_feature_types,
    log_subtitle,
    safe_repr
)

logger = get_logger(__file__)


class ComponentBase(ABC, metaclass=ComponentBaseMeta):
    """Base class for all components."""
    _default_parameters = None

    def __init__(self, parameters=None, component_obj=None, random_seed=0, **kwargs):
        self.random_seed = random_seed
        self._component_obj = component_obj
        self._parameters = parameters or {}
        self._is_fitted = False

    @property
    @classmethod
    @abstractmethod
    def name(cls):
        """Returns string name of this component"""

    @property
    @classmethod
    @abstractmethod
    def model_family(cls):
        """Returns ModelFamily of this component"""

    @classproperty
    def needs_fitting(self):
        """Returns boolean determining if component needs fitting before
            calling predict, predict_proba, transform, or feature_importances.
            This can be overridden to False for components that do not need to be fit
            or whose fit methods do nothing."""
        return True

    @property
    def parameters(self):
        """Returns the parameters which were used to initialize the component"""
        return copy.copy(self._parameters)

    @classproperty
    def default_parameters(cls):
        """Returns the default parameters for this component.

         Our convention is that Component.default_parameters == Component().parameters.

         Returns:
             dict: default parameters for this component.
        """

        if cls._default_parameters is None:
            cls._default_parameters = cls().parameters

        return cls._default_parameters

    def clone(self):
        """Constructs a new component with the same parameters and random state.

        Returns:
            A new instance of this component with identical parameters and random state.
        """
        return self.__class__(**self.parameters, random_seed=self.random_seed)

    def fit(self, X, y=None):
        """Fits component to data

        Arguments:
            X (list, ww.DataTable, pd.DataFrame or np.ndarray): The input training data of shape [n_samples, n_features]
            y (list, ww.DataColumn, pd.Series, np.ndarray, optional): The target training data of length [n_samples]

        Returns:
            self
        """
        X = infer_feature_types(X)
        if y is not None:
            y = infer_feature_types(y)
        try:
            self._component_obj.fit(X, y)
            return self
        except AttributeError:
            raise MethodPropertyNotFoundError("Component requires a fit method or a component_obj that implements fit")

    def describe(self, print_name=False, return_dict=False):
        """Describe a component and its parameters

        Arguments:
            print_name(bool, optional): whether to print name of component
            return_dict(bool, optional): whether to return description as dictionary in the format {"name": name, "parameters": parameters}

        Returns:
            None or dict: prints and returns dictionary
        """
        if print_name:
            title = self.name
            log_subtitle(logger, title)
        for parameter in self.parameters:
            parameter_str = ("\t * {} : {}").format(parameter, self.parameters[parameter])
            logger.info(parameter_str)
        if return_dict:
            component_dict = {"name": self.name}
            component_dict.update({"parameters": self.parameters})
            return component_dict

    def save(self, file_path, pickle_protocol=cloudpickle.DEFAULT_PROTOCOL):
        """Saves component at file path

        Arguments:
            file_path (str): Location to save file
            pickle_protocol (int): The pickle data stream format.

        Returns:
            None
        """
        with open(file_path, 'wb') as f:
            cloudpickle.dump(self, f, protocol=pickle_protocol)

    @staticmethod
    def load(file_path):
        """Loads component at file path

        Arguments:
            file_path (str): Location to load file

        Returns:
            ComponentBase object
        """
        with open(file_path, 'rb') as f:
            return cloudpickle.load(f)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        random_seed_eq = self.random_seed == other.random_seed
        if not random_seed_eq:
            return False
        attributes_to_check = ['_parameters', '_is_fitted']
        for attribute in attributes_to_check:
            if getattr(self, attribute) != getattr(other, attribute):
                return False
        return True

    def __str__(self):
        return self.name

    def __repr__(self):
        parameters_repr = ', '.join([f'{key}={safe_repr(value)}' for key, value in self.parameters.items()])
        return f'{(type(self).__name__)}({parameters_repr})'



from enum import Enum


class ModelFamily(Enum):
    """Enum for family of machine learning models."""

    K_NEIGHBORS = 'k_neighbors'
    """K Nearest Neighbors model family."""

    RANDOM_FOREST = 'random_forest'
    """Random Forest model family."""

    SVM = 'svm'
    """SVM model family."""

    XGBOOST = 'xgboost'
    """XGBoost model family."""

    LIGHTGBM = 'lightgbm'
    """LightGBM model family."""

    LINEAR_MODEL = 'linear_model'
    """Linear model family."""

    CATBOOST = 'catboost'
    """CatBoost model family."""

    EXTRA_TREES = 'extra_trees'
    """Extra Trees model family."""

    ENSEMBLE = 'ensemble'
    """Ensemble model family."""

    DECISION_TREE = 'decision_tree'
    """Decision Tree model family."""

    ARIMA = 'ARIMA'
    """ARIMA model family."""

    BASELINE = 'baseline'
    """Baseline model family."""

    NONE = 'none'
    """None"""

    def __str__(self):
        model_family_dict = {ModelFamily.K_NEIGHBORS.name: "K Nearest Neighbors",
                             ModelFamily.RANDOM_FOREST.name: "Random Forest",
                             ModelFamily.SVM.name: "SVM",
                             ModelFamily.XGBOOST.name: "XGBoost",
                             ModelFamily.LIGHTGBM.name: "LightGBM",
                             ModelFamily.LINEAR_MODEL.name: "Linear",
                             ModelFamily.CATBOOST.name: "CatBoost",
                             ModelFamily.EXTRA_TREES.name: "Extra Trees",
                             ModelFamily.DECISION_TREE.name: "Decision Tree",
                             ModelFamily.BASELINE.name: "Baseline",
                             ModelFamily.ENSEMBLE.name: "Ensemble",
                             ModelFamily.ARIMA.name: "ARIMA",
                             ModelFamily.NONE.name: "None"}
        return model_family_dict[self.name]

    def __repr__(self):
        return "ModelFamily." + self.name

    def is_tree_estimator(self):
        """Checks whether the estimator's model family uses trees."""
        tree_estimators = {self.CATBOOST, self.EXTRA_TREES, self.RANDOM_FOREST,
                           self.DECISION_TREE, self.XGBOOST, self.LIGHTGBM}
        return self in tree_estimators



from .training_validation_split import TrainingValidationSplit
from .time_series_split import TimeSeriesSplit
from .smote_split import KMeansSMOTETVSplit, KMeansSMOTECVSplit
from .smote_tl_split import SMOTETomekTVSplit, SMOTETomekCVSplit
from .rus_split import RandomUnderSamplerCVSplit, RandomUnderSamplerTVSplit
from .smotenc_split import SMOTENCCVSplit, SMOTENCTVSplit
from .balanced_classification_sampler import BalancedClassificationSampler
from .balanced_classification_splitter import BalancedClassificationDataCVSplit, BalancedClassificationDataTVSplit
from .sampler_base import SamplerBase



from bciavm.pipelines.components.transformers import Transformer
from bciavm.utils import infer_feature_types


def _extract_year(col, encode_as_categories=False):
    return col.dt.year, None


_month_to_int_mapping = {"January": 0, "February": 1, "March": 2, "April": 3, "May": 4, "June": 5,
                         "July": 6, "August": 7, "September": 8, "October": 9, "November": 10, "December": 11}


def _extract_month(col, encode_as_categories=False):
    months = col.dt.month_name()
    months_unique = months.unique()
    months_encoded = months.map(lambda m: _month_to_int_mapping[m])
    if encode_as_categories:
        months_encoded = months_encoded.astype("category")
    return months_encoded, {m: _month_to_int_mapping[m] for m in months_unique}


_day_to_int_mapping = {"Sunday": 0, "Monday": 1, "Tuesday": 2, "Wednesday": 3, "Thursday": 4, "Friday": 5,
                       "Saturday": 6}


def _extract_day_of_week(col, encode_as_categories=False):
    days = col.dt.day_name()
    days_unique = days.unique()
    days_encoded = days.map(lambda d: _day_to_int_mapping[d])
    if encode_as_categories:
        days_encoded = days_encoded.astype("category")
    return days_encoded, {d: _day_to_int_mapping[d] for d in days_unique}


def _extract_hour(col, encode_as_categories=False):
    return col.dt.hour, None


class DateTimeFeaturizer(Transformer):
    """Transformer that can automatically featurize DateTime columns."""
    name = "DateTime Featurization Component"
    hyperparameter_ranges = {}
    _function_mappings = {"year": _extract_year,
                          "month": _extract_month,
                          "day_of_week": _extract_day_of_week,
                          "hour": _extract_hour}

    def __init__(self, features_to_extract=None, encode_as_categories=False, random_seed=0, **kwargs):
        """Extracts features from DateTime columns

        Arguments:
            features_to_extract (list): List of features to extract. Valid options include "year", "month", "day_of_week", "hour".
            encode_as_categories (bool): Whether day-of-week and month features should be encoded as pandas "category" dtype.
                This allows OneHotEncoders to encode these features.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        """
        if features_to_extract is None:
            features_to_extract = ["year", "month", "day_of_week", "hour"]
        invalid_features = set(features_to_extract) - set(self._function_mappings.keys())
        if len(invalid_features) > 0:
            raise ValueError("{} are not valid options for features_to_extract".format(", ".join([f"'{feature}'" for feature in invalid_features])))

        parameters = {"features_to_extract": features_to_extract,
                      "encode_as_categories": encode_as_categories}
        parameters.update(kwargs)

        self._date_time_col_names = None
        self._categories = {}
        self.encode_as_categories = encode_as_categories
        super().__init__(parameters=parameters,
                         component_obj=None,
                         random_seed=random_seed)

    def fit(self, X, y=None):
        X = infer_feature_types(X)
        self._date_time_col_names = X.ww.select("datetime").columns.tolist()
        return self

    def transform(self, X, y=None):
        """Transforms data X by creating new features using existing DateTime columns, and then dropping those DateTime columns

        Arguments:
            X (ww.DataTable, pd.DataFrame): Data to transform
            y (ww.DataColumn, pd.Series, optional): Ignored.

        Returns:
            ww.DataTable: Transformed X
        """
        X_ww = infer_feature_types(X)
        features_to_extract = self.parameters["features_to_extract"]
        if len(features_to_extract) == 0:
            return X_ww
        for col_name in self._date_time_col_names:
            for feature in features_to_extract:
                name = f"{col_name}_{feature}"
                features, categories = self._function_mappings[feature](X_ww[col_name], self.encode_as_categories)
                features.name = None
                X_ww.ww[name] = features
                if categories:
                    self._categories[name] = categories
        X_t = X_ww.ww.drop(self._date_time_col_names)
        return X_t

    def get_feature_names(self):
        """Gets the categories of each datetime feature.

        Returns:
           Dictionary, where each key-value pair is a column name and a dictionary
           mapping the unique feature values to their integer encoding.
        """
        return self._categories

    def _get_feature_provenance(self):
        provenance = {}
        for col_name in self._date_time_col_names:
            provenance[col_name] = []
            for feature in self.parameters['features_to_extract']:
                provenance[col_name].append(f'{col_name}_{feature}')
        return provenance



import numpy as np

from bciavm.preprocessing.data_splitters.sampler_base import SamplerBase
from bciavm.utils.woodwork_utils import infer_feature_types


class BalancedClassificationSampler(SamplerBase):
    """Class for balanced classification downsampler."""

    def __init__(self, balanced_ratio=4, min_samples=100, min_percentage=0.1, random_seed=0):
        """
        Arguments:
            balanced_ratio (float): The largest majority:minority ratio that is accepted as 'balanced'. For instance, a 4:1 ratio would be
                represented as 4, while a 6:5 ratio is 1.2. Must be greater than or equal to 1 (or 1:1). Defaults to 4.

            min_samples (int): The minimum number of samples that we must have for any class, pre or post sampling. If a class must be downsampled, it will not be downsampled past this value.
                To determine severe imbalance, the minority class must occur less often than this and must have a class ratio below min_percentage.
                Must be greater than 0. Defaults to 100.

            min_percentage (float): The minimum percentage of the minimum class to total dataset that we tolerate, as long as it is above min_samples.
                To determine severe imbalance, the minority class must have a class ratio below this and must occur less often than min_samples.
                Must be between 0 and 0.5, inclusive. Defaults to 0.1.

            random_seed (int): The seed to use for random sampling. Defaults to 0.
        """
        super().__init__(random_seed=random_seed)
        if balanced_ratio < 1:
            raise ValueError(f"balanced_ratio must be at least 1, but received {balanced_ratio}")
        if min_samples <= 0:
            raise ValueError(f"min_sample must be greater than 0, but received {min_samples}")
        if min_percentage <= 0 or min_percentage > 0.5:
            raise ValueError(f"min_percentage must be between 0 and 0.5, inclusive, but received {min_percentage}")
        self.balanced_ratio = balanced_ratio
        self.min_samples = min_samples
        self.min_percentage = min_percentage
        self.random_state = np.random.RandomState(self.random_seed)

    def _find_ideal_samples(self, y):
        """Returns dictionary of examples to drop for each class if we need to resample.

        Arguments:
            y (pd.Series): Target data passed in

        Returns:
            (dict): dictionary with undersample target class as key, and number of samples to remove as the value.
                If we don't need to resample, returns empty dictionary.
        """
        counts = y.value_counts()
        normalized_counts = y.value_counts(normalize=True)
        class_ratios = normalized_counts / min(normalized_counts)
        # if no class ratios are greater than what we consider balanced, then the target is balanced
        if all(class_ratios <= self.balanced_ratio):
            return {}
        # if any classes have less than min_samples counts and are less than min_percentage of the total data,
        # then it's severely imbalanced
        if any(counts < self.min_samples) and any(normalized_counts < self.min_percentage):
            return {}
        # otherwise, we are imbalanced enough to perform on this
        undersample_classes = counts[class_ratios > self.balanced_ratio].index.values
        # find goal size, round it down if it's a float
        minority_class = min(counts.values)
        goal_value = max(int((self.balanced_ratio * minority_class) // 1), self.min_samples)
        # we don't want to drop less than 0 rows
        drop_values = {k: max(0, counts[k] - goal_value) for k in undersample_classes}
        return {k: v for k, v in drop_values.items() if v > 0}

    def fit_resample(self, X, y):
        """Resampling technique for this sampler.

        Arguments:
            X (pd.DataFrame): Training data to fit and resample
            y (pd.Series): Training data targets to fit and resample

        Returns:
            list: Indices to keep for training data
        """
        y = infer_feature_types(y)
        result = self._find_ideal_samples(y)
        indices_to_drop = []
        if len(result):
            # iterate through the classes we need to undersample and remove the number of samples we need to remove
            for key, value in result.items():
                indices = y.index[y == key].values
                indices_to_remove = self.random_state.choice(indices, value, replace=False)
                indices_to_drop.extend(indices_to_remove)
        # indices of the y datacolumn
        original_indices = list(set(y.index.values).difference(set(indices_to_drop)))
        return original_indices



from bciavm.pipelines.components.transformers import Transformer
from bciavm.utils.logger import get_logger

logger = get_logger(__file__)


class TextTransformer(Transformer):
    """Base class for all transformers working with text features"""

    def __init__(self, text_columns=['description', 'thumbnail'], component_obj=None, random_state=0, **kwargs):
        """Creates a transformer to perform TF-IDF transformation and Singular Value Decomposition for text columns.

        Arguments:
            text_columns (list): list of feature names which should be treated as text features.
            random_state (int, np.random.RandomState): Seed for the random number generator.
        """
        parameters = {'text_columns': text_columns}
        parameters.update(kwargs)

        self._all_text_columns = text_columns or []
        super().__init__(parameters=parameters,
                         component_obj=component_obj,
                         random_state=random_state)

    def _get_text_columns(self, X):
        """Returns the ordered list of columns names in the input which have been designated as text columns."""
        columns = []
        missing_columns = []
        for col_name in self._all_text_columns:
            if col_name in X.columns:
                columns.append(col_name)
            else:
                missing_columns.append(col_name)
        if len(columns) == 0:
            raise AttributeError("None of the provided text column names match the columns in the given DataFrame")
        if len(columns) < len(self._all_text_columns):
            logger.warn("Columns {} were not found in the given DataFrame, ignoring".format(missing_columns))
        return columns



from .datetime_featurizer import DateTimeFeaturizer
from .drop_null_columns import DropNullColumns
from .text_transformer import TextTransformer
from .lsa import LSA
from .text_featurizer import TextFeaturizer
from .delayed_feature_transformer import DelayedFeatureTransformer
from .featuretools import DFSTransformer
from .polynomial_detrender import PolynomialDetrender
from .preprocessor import PreprocessTransformer



from .transformer import Transformer
from .encoders import OneHotEncoder, TargetEncoder
from .feature_selection import FeatureSelector, RFClassifierSelectFromModel, RFRegressorSelectFromModel
from .imputers import PerColumnImputer, SimpleImputer, Imputer
from .scalers import StandardScaler
from .column_selectors import DropColumns, SelectColumns
from .dimensionality_reduction import LinearDiscriminantAnalysis, PCA
from .preprocessing import DateTimeFeaturizer, DropNullColumns, LSA, TextFeaturizer, DelayedFeatureTransformer, DFSTransformer, PolynomialDetrender, PreprocessTransformer


from sklearn.ensemble import StackingRegressor
from sklearn.model_selection import KFold

from bciavm.model_family import ModelFamily
from bciavm.pipelines.components import LinearRegressor
from bciavm.pipelines.components.ensemble import StackedEnsembleBase
from bciavm.problem_types import ProblemTypes


class StackedEnsembleRegressor(StackedEnsembleBase):
    """Stacked Ensemble Regressor."""
    name = "Stacked Ensemble Regressor"
    model_family = ModelFamily.ENSEMBLE
    supported_problem_types = [ProblemTypes.REGRESSION, ProblemTypes.TIME_SERIES_REGRESSION]
    hyperparameter_ranges = {}
    _stacking_estimator_class = StackingRegressor
    _default_final_estimator = LinearRegressor
    _default_cv = KFold

    def __init__(self, input_pipelines=None, final_estimator=None,
                 cv=None, n_jobs=-1, random_seed=0, **kwargs):
        """Stacked ensemble regressor.

        Arguments:
            input_pipelines (list(PipelineBase or subclass obj)): List of pipeline instances to use as the base estimators.
                This must not be None or an empty list or else EnsembleMissingPipelinesError will be raised.
            final_estimator (Estimator or subclass): The regressor used to combine the base estimators. If None, uses LinearRegressor.
            cv (int, cross-validation generator or an iterable): Determines the cross-validation splitting strategy used to train final_estimator.
                For int/None inputs, KFold is used. Defaults to None.
                Possible inputs for cv are:

                - None: 3-fold cross validation
                - int: the number of folds in a (Stratified) KFold
                - An scikit-learn cross-validation generator object
                - An iterable yielding (train, test) splits
            n_jobs (int or None): Non-negative integer describing level of parallelism used for pipelines.
                None and 1 are equivalent. If set to -1, all CPUs are used. For n_jobs below -1, (n_cpus + 1 + n_jobs) are used.
                Defaults to None.
                - Note: there could be some multi-process errors thrown for values of `n_jobs != 1`. If this is the case, please use `n_jobs = 1`.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        """
        super().__init__(input_pipelines=input_pipelines, final_estimator=final_estimator, cv=cv,
                         n_jobs=n_jobs, random_seed=random_seed, **kwargs)





from abc import ABCMeta
from functools import wraps


class BaseMeta(ABCMeta):
    """Metaclass that overrides creating a new component or pipeline by wrapping methods with validators and setters"""

    FIT_METHODS = ['fit', 'fit_transform']
    METHODS_TO_CHECK = ['predict', 'predict_proba', 'transform', 'inverse_transform']
    PROPERTIES_TO_CHECK = ['feature_importance']

    @classmethod
    def set_fit(cls, method):
        @wraps(method)
        def _set_fit(self, X, y=None):
            return_value = method(self, X, y)
            self._is_fitted = True
            return return_value
        return _set_fit

    def __new__(cls, name, bases, dct):
        for attribute in dct:
            if attribute in cls.FIT_METHODS:
                dct[attribute] = cls.set_fit(dct[attribute])
            if attribute in cls.METHODS_TO_CHECK:
                dct[attribute] = cls.check_for_fit(dct[attribute])
            if attribute in cls.PROPERTIES_TO_CHECK:
                property_orig = dct[attribute]
                dct[attribute] = property(cls.check_for_fit(property_orig.__get__),
                                          property_orig.__set__,
                                          property_orig.__delattr__,
                                          property_orig.__doc__)
        return super().__new__(cls, name, bases, dct)



from abc import abstractmethod

from bciavm.exceptions import MethodPropertyNotFoundError
from bciavm.model_family import ModelFamily
from bciavm.pipelines.components import ComponentBase
from bciavm.utils import infer_feature_types


class Estimator(ComponentBase):
    """A component that fits and predicts given data.

    To implement a new Estimator, define your own class which is a subclass of Estimator, including
    a name and a list of acceptable ranges for any parameters to be tuned during the automl search (hyperparameters).
    Define an `__init__` method which sets up any necessary state and objects. Make sure your `__init__` only
    uses standard keyword arguments and calls `super().__init__()` with a parameters dict. You may also override the
    `fit`, `transform`, `fit_transform` and other methods in this class if appropriate.

    To see some examples, check out the definitions of any Estimator component.
    """
    # We can't use the inspect module to dynamically determine this because of issue 1582
    predict_uses_y = False
    model_family = ModelFamily.NONE

    @property
    @classmethod
    @abstractmethod
    def supported_problem_types(cls):
        """Problem types this estimator supports"""

    def __init__(self, parameters=None, component_obj=None, random_seed=0, **kwargs):
        self.input_feature_names = None
        super().__init__(parameters=parameters, component_obj=component_obj, random_seed=random_seed, **kwargs)

    def _manage_woodwork(self, X, y=None):
        """Function to convert the input and target data to Pandas data structures."""
        if X is not None:
            X = infer_feature_types(X)
        if y is not None:
            y = infer_feature_types(y)
        return X, y

    def fit(self, X, y=None):
        X, y = self._manage_woodwork(X, y)
        self.input_feature_names = list(X.columns)
        self._component_obj.fit(X, y)
        return self

    def predict(self, X):
        """Make predictions using selected features.

        Arguments:
            X (ww.DataTable, pd.DataFrame, or np.ndarray): Data of shape [n_samples, n_features]

        Returns:
            ww.DataColumn: Predicted values
        """
        try:
            X = infer_feature_types(X)
            predictions = self._component_obj.predict(X)
        except AttributeError:
            raise MethodPropertyNotFoundError("Estimator requires a predict method or a component_obj that implements predict")
        return infer_feature_types(predictions)

    def predict_proba(self, X):
        """Make probability estimates for labels.

        Arguments:
            X (ww.DataTable, pd.DataFrame, or np.ndarray): Features

        Returns:
            ww.DataTable: Probability estimates
        """
        try:
            X = infer_feature_types(X)
            pred_proba = self._component_obj.predict_proba(X)
        except AttributeError:
            raise MethodPropertyNotFoundError("Estimator requires a predict_proba method or a component_obj that implements predict_proba")
        return infer_feature_types(pred_proba)

    @property
    def feature_importance(self):
        """Returns importance associated with each feature.

        Returns:
            np.ndarray: Importance associated with each feature
        """
        try:
            return self._component_obj.feature_importances_
        except AttributeError:
            raise MethodPropertyNotFoundError("Estimator requires a feature_importance property or a component_obj that implements feature_importances_")

    def __eq__(self, other):
        return super().__eq__(other) and self.supported_problem_types == other.supported_problem_types



class MethodPropertyNotFoundError(Exception):
    """Exception to raise when a class is does not have an expected method or property."""
    pass


class PipelineNotFoundError(Exception):
    """An exception raised when a particular pipeline is not found in automl search results"""
    pass


class ObjectiveNotFoundError(Exception):
    """Exception to raise when specified objective does not exist."""
    pass


class IllFormattedClassNameError(Exception):
    """Exception to raise when a class name does not comply with bciavm standards"""
    pass


class MissingComponentError(Exception):
    """An exception raised when a component is not found in all_components()"""
    pass


class ComponentNotYetFittedError(Exception):
    """An exception to be raised when predict/predict_proba/transform is called on a component without fitting first."""
    pass


class PipelineNotYetFittedError(Exception):
    """An exception to be raised when predict/predict_proba/transform is called on a pipeline without fitting first."""
    pass


class AutoMLSearchException(Exception):
    """Exception raised when all pipelines in an automl batch return a score of NaN for the primary objective."""
    pass


class EnsembleMissingPipelinesError(Exception):
    """An exception raised when an ensemble is missing `estimators` (list) as a parameter."""
    pass


class PipelineScoreError(Exception):
    """An exception raised when a pipeline errors while scoring any objective in a list of objectives.

    Arguments:
        exceptions (dict): A dictionary mapping an objective name (str) to a tuple of the form (exception, traceback).
            All of the objectives that errored will be stored here.
        scored_successfully (dict): A dictionary mapping an objective name (str) to a score value. All of the objectives
            that did not error will be stored here.
    """

    def __init__(self, exceptions, scored_successfully):
        self.exceptions = exceptions
        self.scored_successfully = scored_successfully

        # Format the traceback message
        exception_list = []
        for objective, (exception, tb) in exceptions.items():
            exception_list.append(f"{objective} encountered {str(exception.__class__.__name__)} with message ({str(exception)}):\n")
            exception_list.extend(tb)
        message = "\n".join(exception_list)

        self.message = message
        super().__init__(message)


class DataCheckInitError(Exception):
    """Exception raised when a data check can't initialize with the parameters given."""


class NullsInColumnWarning(UserWarning):
    """Warning thrown when there are null values in the column of interest"""


class ObjectiveCreationError(Exception):
    """Exception when get_objective tries to instantiate an objective and required args are not provided."""




from .exceptions import (
    MethodPropertyNotFoundError,
    PipelineNotFoundError,
    ObjectiveNotFoundError,
    IllFormattedClassNameError,
    MissingComponentError,
    ComponentNotYetFittedError,
    PipelineNotYetFittedError,
    AutoMLSearchException,
    PipelineScoreError,
    DataCheckInitError,
    EnsembleMissingPipelinesError,
    NullsInColumnWarning,
    ObjectiveCreationError
)



from sklearn.model_selection import StratifiedKFold

from bciavm.preprocessing.data_splitters.balanced_classification_sampler import (
    BalancedClassificationSampler
)
from bciavm.preprocessing.data_splitters.base_splitters import (
    BaseUnderSamplingSplitter
)
from bciavm.preprocessing.data_splitters.training_validation_split import (
    TrainingValidationSplit
)


class BalancedClassificationDataTVSplit(BaseUnderSamplingSplitter):
    """Data splitter for generating training and validation split using Balanced Classification Data Sampler."""

    def __init__(self, balanced_ratio=4, min_samples=100, min_percentage=0.1, test_size=0.25, shuffle=True, random_seed=0):
        """Create Balanced Classification Data TV splitter

        Arguments:
            balanced_ratio (float): The largest majority:minority ratio that is accepted as 'balanced'. For instance, a 4:1 ratio would be
                represented as 4, while a 6:5 ratio is 1.2. Must be greater than or equal to 1 (or 1:1). Defaults to 4.

            min_samples (int): The minimum number of samples that we must have for any class, pre or post sampling. If a class must be downsampled, it will not be downsampled past this value.
                To determine severe imbalance, the minority class must occur less often than this and must have a class ratio below min_percentage.
                Must be greater than 0. Defaults to 100.

            min_percentage (float): The minimum percentage of the minimum class to total dataset that we tolerate, as long as it is above min_samples.
                If min_percentage and min_samples are not met, treat this as severely imbalanced, and we will not resample the data.
                Must be between 0 and 0.5, inclusive. Defaults to 0.1.

            test_size (float): The size of the test split. Defaults to 0.25.

            shuffle (bool): Whether or not to shuffle the data before splitting. Defaults to True.

            random_seed (int): The seed to use for random sampling. Defaults to 0.
        """
        self.sampler = BalancedClassificationSampler(balanced_ratio=balanced_ratio, min_samples=min_samples, min_percentage=min_percentage, random_seed=random_seed)
        super().__init__(sampler=self.sampler, n_splits=1, random_seed=random_seed)
        self.shuffle = shuffle
        self.test_size = test_size
        self.balanced_ratio = balanced_ratio
        self.min_samples = min_samples
        self.min_percentage = min_percentage
        self.splitter = TrainingValidationSplit(test_size=test_size, shuffle=shuffle, random_seed=random_seed)


class BalancedClassificationDataCVSplit(BaseUnderSamplingSplitter):
    """Data splitter for generating k-fold cross-validation split using Balanced Classification Data Sampler."""

    def __init__(self, balanced_ratio=4, min_samples=100, min_percentage=0.1, n_splits=3, shuffle=True, random_seed=0):
        """Create Balanced Classification Data CV splitter

        Arguments:
            balanced_ratio (float): The largest majority:minority ratio that is accepted as 'balanced'. For instance, a 4:1 ratio would be
                represented as 4, while a 6:5 ratio is 1.2. Must be greater than or equal to 1 (or 1:1). Defaults to 4.

            min_samples (int): The minimum number of samples that we must have for any class, pre or post sampling. If a class must be downsampled, it will not be downsampled past this value.
                To determine severe imbalance, the minority class must occur less often than this and must have a class ratio below min_percentage.
                Must be greater than 0. Defaults to 100.

            min_percentage (float): The minimum percentage of the minimum class to total dataset that we tolerate, as long as it is above min_samples.
                If min_percentage and min_samples are not met, treat this as severely imbalanced, and we will not resample the data.
                Must be between 0 and 0.5, inclusive. Defaults to 0.1.

            n_splits (int): The number of splits to use for cross validation. Defaults to 3.

            shuffle (bool): Whether or not to shuffle the data before splitting. Defaults to True.

            random_seed (int): The seed to use for random sampling. Defaults to 0.
        """
        self.sampler = BalancedClassificationSampler(balanced_ratio=balanced_ratio, min_samples=min_samples, min_percentage=min_percentage, random_seed=random_seed)
        super().__init__(sampler=self.sampler, n_splits=n_splits, random_seed=random_seed)
        self.shuffle = shuffle
        self.balanced_ratio = balanced_ratio
        self.min_samples = min_samples
        self.min_percentage = min_percentage
        self.splitter = StratifiedKFold(n_splits=n_splits, shuffle=shuffle, random_state=random_seed)




import numpy as np
import pandas as pd
import woodwork as ww

from bciavm import Boolean, Integer
from bciavm.utils.gen_utils import is_all_numeric

numeric_and_boolean_ww = [Integer, ww.logical_types.Double, Boolean]


def _numpy_to_pandas(array):
    if len(array.shape) == 1:
        data = pd.Series(array)
    else:
        data = pd.DataFrame(array)
    return data


def _list_to_pandas(list):
    return _numpy_to_pandas(np.array(list))


def infer_feature_types(data, feature_types=None):
    """Create a Woodwork structure from the given list, pandas, or numpy input, with specified types for columns.
        If a column's type is not specified, it will be inferred by Woodwork.
    Arguments:
        data (pd.DataFrame, pd.Series): Input data to convert to a Woodwork data structure.
        feature_types (string, ww.logical_type obj, dict, optional): If data is a 2D structure, feature_types must be a dictionary
            mapping column names to the type of data represented in the column. If data is a 1D structure, then feature_types must be
            a Woodwork logical type or a string representing a Woodwork logical type ("Double", "Integer", "Boolean", "Categorical", "Datetime", "NaturalLanguage")
    Returns:
        A Woodwork data structure where the data type of each column was either specified or inferred.
    """
    if isinstance(data, list):
        data = _list_to_pandas(data)
    elif isinstance(data, np.ndarray):
        data = _numpy_to_pandas(data)

    feature_t = None
    try:
        feature_t = {}
        for col in data.columns:
            if col in feature_types:
                feature_t[col] = feature_types[col]
    except: pass

    ww_data = data.copy()

    if isinstance(data, pd.Series):
        if data.ww._schema is not None:
            ww_data = ww.init_series(ww_data, logical_type=data.ww.logical_type)
        else:
            ww_data = ww.init_series(ww_data, logical_type=feature_t)
    else:
        if data.ww.schema is not None:
            nullable_types = {ww.logical_types.Boolean, ww.logical_types.Integer}
            # if set(data.ww.logical_types.values()).intersection(nullable_types):
            #     raise ValueError()
            ww_data.ww.init(logical_types=data.ww.logical_types,
                            semantic_tags=data.ww.semantic_tags)
        else:
            ww_data.ww.init(logical_types=feature_t)

    return ww_data


def _retain_custom_types_and_initalize_woodwork(old_logical_types, new_dataframe, ltypes_to_ignore=None):
    """
    Helper method which will take an old Woodwork DataTable and a new pandas DataFrame and return a
    new DataTable that will try to retain as many logical types from the old DataTable that exist in the new
    pandas DataFrame as possible.
    Arguments:
        old_datatable (ww.DataTable): Woodwork DataTable to use
        new_dataframe (pd.DataFrame): Pandas data structure
        ltypes_to_ignore (list): List of Woodwork logical types to ignore. Columns from the old DataTable that have a logical type
        specified in this list will not have their logical types carried over to the new DataTable returned
    Returns:
        A new DataTable where any of the columns that exist in the old input DataTable and the new DataFrame try to retain
        the original logical type, if possible and not specified to be ignored.
    """
    # if ltypes_to_ignore is None:
    #     ltypes_to_ignore = []
    # col_intersection = set(old_logical_types.keys()).intersection(set(new_dataframe.columns))
    # retained_logical_types = {col: ltype for col, ltype in old_logical_types.items() if col in col_intersection and ltype not in ltypes_to_ignore}
    # new_dataframe.ww.init(logical_types=retained_logical_types)
    new_dataframe = infer_feature_types(new_dataframe)
    return new_dataframe


def _convert_numeric_dataset_pandas(X, y):
    """Convert numeric and non-null data to pandas datatype. Raises ValueError if there is null or non-numeric data.
    Used with data sampler strategies.
    Arguments:
        X (pd.DataFrame, np.ndarray, ww.DataTable): Data to transform
        y (pd.Series, np.ndarray, ww.DataColumn): Target data
    Returns:
        Tuple(pd.DataFrame, pd.Series): Transformed X and y"""
    X_ww = infer_feature_types(X)
    if not is_all_numeric(X_ww):
        raise ValueError('Values not all numeric or there are null values provided in the dataset')
    y_ww = infer_feature_types(y)
    return X_ww, y_ww



import pandas as pd

from bciavm.exceptions import MethodPropertyNotFoundError
from bciavm.model_family import ModelFamily
from bciavm.pipelines.components import ComponentBase
from bciavm.utils import (
    _retain_custom_types_and_initalize_woodwork,
    infer_feature_types
)


class Transformer(ComponentBase):
    """A component that may or may not need fitting that transforms data.
    These components are used before an estimator.

    To implement a new Transformer, define your own class which is a subclass of Transformer, including
    a name and a list of acceptable ranges for any parameters to be tuned during the automl search (hyperparameters).
    Define an `__init__` method which sets up any necessary state and objects. Make sure your `__init__` only
    uses standard keyword arguments and calls `super().__init__()` with a parameters dict. You may also override the
    `fit`, `transform`, `fit_transform` and other methods in this class if appropriate.

    To see some examples, check out the definitions of any Transformer component.
    """

    model_family = ModelFamily.NONE

    def transform(self, X, y=None):
        """Transforms data X.

        Arguments:
            X (ww.DataTable, pd.DataFrame): Data to transform.
            y (ww.DataColumn, pd.Series, optional): Target data.

        Returns:
            ww.DataTable: Transformed X
        """
        X_ww = infer_feature_types(X)
        if y is not None:
            y = infer_feature_types(y)
        try:
            X_t = self._component_obj.transform(X, y)
        except AttributeError:
            raise MethodPropertyNotFoundError("Transformer requires a transform method or a component_obj that implements transform")
        X_t_df = pd.DataFrame(X_t, columns=X_ww.columns, index=X_ww.index)
        return _retain_custom_types_and_initalize_woodwork(X_ww.ww.logical_types, X_t_df)

    def fit_transform(self, X, y=None):
        """Fits on X and transforms X

        Arguments:
            X (ww.DataTable, pd.DataFrame): Data to fit and transform
            y (ww.DataColumn, pd.Series): Target data

        Returns:
            ww.DataTable: Transformed X
        """
        X_ww = infer_feature_types(X)
        if y is not None:
            y_ww = infer_feature_types(y)
        try:
            X_t = self._component_obj.fit_transform(X_ww, y_ww)
            return _retain_custom_types_and_initalize_woodwork(X_ww.ww.logical_types, X_t)
        except AttributeError:
            try:
                return self.fit(X, y).transform(X, y)
            except MethodPropertyNotFoundError as e:
                raise e

    def _get_feature_provenance(self):
        return {}



from bciavm.exceptions import EnsembleMissingPipelinesError
from bciavm.model_family import ModelFamily
from bciavm.pipelines.components import Estimator
from bciavm.pipelines.components.utils import scikit_learn_wrapped_estimator
from bciavm.utils import classproperty

_nonstackable_model_families = [ModelFamily.BASELINE, ModelFamily.NONE]


class StackedEnsembleBase(Estimator):
    """Stacked Ensemble Base Class."""
    model_family = ModelFamily.ENSEMBLE
    _stacking_estimator_class = None
    _default_final_estimator = None
    _default_cv = None

    def __init__(self, input_pipelines=None, final_estimator=None, cv=None, n_jobs=None, random_seed=0, **kwargs):
        """Stacked ensemble base class.

        Arguments:
            input_pipelines (list(PipelineBase or subclass obj)): List of pipeline instances to use as the base estimators.
                This must not be None or an empty list or else EnsembleMissingPipelinesError will be raised.
            final_estimator (Estimator or subclass): The estimator used to combine the base estimators.
            cv (int, cross-validation generator or an iterable): Determines the cross-validation splitting strategy used to train final_estimator.
                For int/None inputs, if the estimator is a classifier and y is either binary or multiclass, StratifiedKFold is used. In all other cases, KFold is used.
                Possible inputs for cv are:

                - None: 5-fold cross validation
                - int: the number of folds in a (Stratified) KFold
                - An scikit-learn cross-validation generator object
                - An iterable yielding (train, test) splits
            n_jobs (int or None): Non-negative integer describing level of parallelism used for pipelines.
                None and 1 are equivalent. If set to -1, all CPUs are used. For n_jobs below -1, (n_cpus + 1 + n_jobs) are used.
                Defaults to None.
                - Note: there could be some multi-process errors thrown for values of `n_jobs != 1`. If this is the case, please use `n_jobs = 1`.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        """
        if not input_pipelines:
            raise EnsembleMissingPipelinesError("`input_pipelines` must not be None or an empty list.")
        if [pipeline for pipeline in input_pipelines if pipeline.model_family in _nonstackable_model_families]:
            raise ValueError("Pipelines with any of the following model families cannot be used as base pipelines: {}".format(_nonstackable_model_families))

        parameters = {
            "input_pipelines": input_pipelines,
            "final_estimator": final_estimator,
            "cv": cv,
            "n_jobs": n_jobs
        }
        parameters.update(kwargs)

        if len(set([pipeline.problem_type for pipeline in input_pipelines])) > 1:
            raise ValueError("All pipelines must have the same problem type.")

        cv = cv or self._default_cv(n_splits=3, random_state=random_seed, shuffle=True)
        estimators = [scikit_learn_wrapped_estimator(pipeline) for pipeline in input_pipelines]
        final_estimator = scikit_learn_wrapped_estimator(final_estimator or self._default_final_estimator())
        sklearn_parameters = {
            "estimators": [(f"({idx})", estimator) for idx, estimator in enumerate(estimators)],
            "final_estimator": final_estimator,
            "cv": cv,
            "n_jobs": n_jobs
        }
        sklearn_parameters.update(kwargs)
        super().__init__(parameters=parameters,
                         component_obj=self._stacking_estimator_class(**sklearn_parameters),
                         random_seed=random_seed)

    @property
    def feature_importance(self):
        """Not implemented for StackedEnsembleClassifier and StackedEnsembleRegressor"""
        raise NotImplementedError("feature_importance is not implemented for StackedEnsembleClassifier and StackedEnsembleRegressor")

    @classproperty
    def default_parameters(cls):
        """Returns the default parameters for stacked ensemble classes.

         Returns:
             dict: default parameters for this component.
        """
        return {
            'final_estimator': None,
            'cv': None,
            'n_jobs': -1,
        }



from skopt.space import Integer, Real

from bciavm.model_family import ModelFamily
from bciavm.pipelines.components.estimators import Estimator
from bciavm.problem_types import ProblemTypes
from bciavm.utils.gen_utils import (
    _rename_column_names_to_numeric,
    import_or_raise
)

import time, sys, random, json
import hyperopt
from hyperopt import fmin, tpe, hp, SparkTrials, Trials, STATUS_OK
from hyperopt.pyll import scope
from math import exp
from dask_ml.wrappers import ParallelPostFit


class XGBoostRegressor(Estimator):
    """XGBoost Regressor."""
    name = "XGBoost Regressor"
    hyperparameter_ranges = {
        "eta": Real(0.000001, 1),
        "max_depth": Integer(1, 20),
        "min_child_weight": Real(1, 10),
        "n_estimators": Integer(1, 1000),
    }

    search_space = {
        'max_depth' : scope.int(hp.quniform('max_depth', 4, 100, 1)),
        'learning_rate' : hp.loguniform('learning_rate', -3, 0),
        'reg_alpha' : hp.loguniform('reg_alpha', -5, -1),
        'reg_lambda' : hp.loguniform('reg_lambda', -6, -1),
        'min_child_weight' : hp.loguniform('min_child_weight', -1, 3),
        'n_estimators' : hp.choice('n_estimators', [ x for x in range(100, 1000) ]),
        'objective' : 'reg:squarederror',
        'metric' : 'mae',
        'seed' : 123,  # Set a seed for deterministic training
    }

    params = {
        'metric': 'mae',
        'objective': 'reg:squarederror',
        'verbose_eval': True,
        'seed': 123,
        'reg_alpha': 0.043706006022706405,
        'max_depth': 14,
        'learning_rate': 0.06325261812661621,
        'min_child_weight': 0.6718934260322275,
        'reg_lambda': 0.026408282583277758
    }
    model_family = ModelFamily.XGBOOST
    supported_problem_types = [ProblemTypes.REGRESSION, ProblemTypes.TIME_SERIES_REGRESSION]

    # xgboost supports seeds from -2**31 to 2**31 - 1 inclusive. these limits ensure the random seed generated below
    # is within that range.
    SEED_MIN = -2**31
    SEED_MAX = 2**31 - 1

    def __init__(self, learning_rate=0.06325261812661621,
                        max_depth=14,
                        min_child_weight=0.6718934260322275,
                        n_estimators=766,
                        reg_alpha=0.043706006022706405,
                        reg_lambda=0.026408282583277758,
                        random_seed=0,
                        **kwargs):

        parameters = {"learning_rate": learning_rate,
                      "max_depth": max_depth,
                      "min_child_weight": min_child_weight,
                      "reg_alpha": reg_alpha,
                      "reg_lambda": reg_lambda,
                      "n_estimators": n_estimators}

        parameters.update(kwargs)

        xgb_error_msg = "XGBoost is not installed. Please install using `pip install xgboost.`"
        xgb = import_or_raise("xgboost", error_msg=xgb_error_msg)
        xgb_Regressor = xgb.XGBRegressor(n_jobs=4, random_state=random_seed,
                                         **parameters)
        super().__init__(parameters=parameters,
                         component_obj=xgb_Regressor,
                         random_seed=random_seed)

    def fit(self, X, y=None):
        X, y = super()._manage_woodwork(X, y)
        X = self._ensure_types(X)
        self.input_feature_names = list(X.columns)
        X = _rename_column_names_to_numeric(X, flatten_tuples=False)
        self._component_obj.fit(X, y)
        return self

    def _ensure_types(self, X):
        for col in X.columns:
            try: X[col] = X[col].astype(float)
            except: X = X.drop(col, axis=1)
            if col in ['unit_indx','_c0']:
                try: X = X.drop(col, axis=1)
                except: pass
        return X

    def predict(self, X):
        X = self._ensure_types(X)
        X = X[self.input_feature_names]
        X = _rename_column_names_to_numeric(X, flatten_tuples=False)
        return super().predict(X)

    @property
    def feature_importance(self):
        return self._component_obj.feature_importances_



import copy

import pandas as pd
from sklearn.preprocessing import LabelEncoder, OrdinalEncoder
from woodwork import logical_types

from bciavm.pipelines.components.transformers.transformer import Transformer
from bciavm.utils import (
    _retain_custom_types_and_initalize_woodwork,
    infer_feature_types
)


class DelayedFeatureTransformer(Transformer):
    """Transformer that delayes input features and target variable for time series problems."""
    name = "Delayed Feature Transformer"
    hyperparameter_ranges = {}
    needs_fitting = False

    def __init__(self, max_delay=2, delay_features=True, delay_target=True, gap=1, random_seed=0, **kwargs):
        """Creates a DelayedFeatureTransformer.

        Arguments:
            max_delay (int): Maximum number of time units to delay each feature.
            delay_features (bool): Whether to delay the input features.
            delay_target (bool): Whether to delay the target.
            gap (int): The number of time units between when the features are collected and
                when the target is collected. For example, if you are predicting the next time step's target, gap=1.
                This is only needed because when gap=0, we need to be sure to start the lagging of the target variable
                at 1.
            random_seed (int): Seed for the random number generator. This transformer performs the same regardless of the random seed provided.
        """
        self.max_delay = max_delay
        self.delay_features = delay_features
        self.delay_target = delay_target

        # If 0, start at 1
        self.start_delay_for_target = int(gap == 0)

        parameters = {"max_delay": max_delay, "delay_target": delay_target, "delay_features": delay_features,
                      "gap": gap}
        parameters.update(kwargs)
        super().__init__(parameters=parameters, random_seed=random_seed)

    def fit(self, X, y=None):
        """Fits the DelayFeatureTransformer.

        Arguments:
            X (ww.DataTable, pd.DataFrame or np.ndarray): The input training data of shape [n_samples, n_features]
            y (ww.DataColumn, pd.Series, optional): The target training data of length [n_samples]

        Returns:
            self
        """
        return self

    @staticmethod
    def _encode_y_while_preserving_index(y):
        y_encoded = LabelEncoder().fit_transform(y)
        y = pd.Series(y_encoded, index=y.index)
        return y

    @staticmethod
    def _get_categorical_columns(X):
        return [name for name, column in X.ww.columns.items() if column.logical_type == logical_types.Categorical]

    @staticmethod
    def _encode_X_while_preserving_index(X_categorical):
        return pd.DataFrame(OrdinalEncoder().fit_transform(X_categorical),
                            columns=X_categorical.columns, index=X_categorical.index)

    def transform(self, X, y=None):
        """Computes the delayed features for all features in X and y.

        For each feature in X, it will add a column to the output dataframe for each
        delay in the (inclusive) range [1, max_delay]. The values of each delayed feature are simply the original
        feature shifted forward in time by the delay amount. For example, a delay of 3 units means that the feature
        value at row n will be taken from the n-3rd row of that feature

        If y is not None, it will also compute the delayed values for the target variable.

        Arguments:
            X (ww.DataTable, pd.DataFrame or None): Data to transform. None is expected when only the target variable is being used.
            y (ww.DataColumn, pd.Series, or None): Target.

        Returns:
            ww.DataTable: Transformed X.
        """
        if X is None:
            X = pd.DataFrame()
        # Normalize the data into pandas objects
        X_ww = infer_feature_types(X)
        original_logical_types = copy.deepcopy(X_ww.ww.logical_types)
        categorical_columns = self._get_categorical_columns(X_ww)
        if self.delay_features and len(X) > 0:
            X_categorical = self._encode_X_while_preserving_index(X_ww[categorical_columns])
            for col_name in X_ww:
                col = X_ww[col_name]
                if col_name in categorical_columns:
                    col = X_categorical[col_name]
                X_ww = X_ww.assign(**{f"{col_name}_delay_{t}": col.shift(t) for t in range(1, self.max_delay + 1)})
        # Handle cases where the target was passed in
        if self.delay_target and y is not None:
            y = infer_feature_types(y)
            if y.ww.logical_type == logical_types.Categorical:
                y = self._encode_y_while_preserving_index(y)
            X_ww = X_ww.assign(**{f"target_delay_{t}": y.shift(t)
                                  for t in range(self.start_delay_for_target, self.max_delay + 1)})

        return _retain_custom_types_and_initalize_woodwork(original_logical_types, X_ww)

    def fit_transform(self, X, y):
        return self.fit(X, y).transform(X, y)
